__all__ = ('CommandContext', )

from ...backend.export import export

from .command_helpers import get_command_category_trace, handle_exception, run_checks
from .exceptions import CommandCheckError
from .responding import process_command_coroutine


@export
class CommandContext(object):
    """
    Represents a command context within the command is invoked.
    
    Attributes
    ----------
    client : ``Client``
        The client who received the message.
    command : ``Command``
        The command to invoke.
    command_category_trace : `None` or `tuple` of ``CommandCategory``
        Trace of command categories till to the command to invoke if applicable.
    command_function : `None` or ``CommandFunction``
        The command's function to run.
    command_keyword_parameters : `dict` of (`str`, `Any`) items
        Keyword parameters to pass to the command function.
    command_positional_parameters : `list` of `Any`
        Positional parameters to pass to the command function.
    content : `str`
        The message's content after prefix.
    message : ``Message``
        The received message.
    parameters : `None` or `dict` of (`str`, `Any`)
        The parsed parameters.
    prefix : `str`
        The matched prefix or the client's prefix for the given respective message.
    """
    __slots__ = ('client', 'command', 'command_category_trace', 'command_function', 'command_keyword_parameters',
        'command_positional_parameters', 'content', 'message', 'parameters', 'prefix')
    
    def __new__(cls, client, message, prefix, content, command):
        """
        Creates a new command context instance.
        
        client : ``Client``
            The client who received the message.
        message : ``Message``
            The received message.
        prefix : `str`
            The matched prefix or the client's prefix for the given respective message.
        content : `str`
            The message's content after prefix.
        command : ``Command``
            The command to invoke.
        """
        self = object.__new__(cls)
        self.client = client
        self.message = message
        self.parameters = None
        self.prefix = prefix
        self.content = content
        self.command = command
        self.command_category_trace = None
        self.command_function = None
        self.command_keyword_parameters = {}
        self.command_positional_parameters = []
        return self
    
    def __repr__(self):
        """Returns the context's representation."""
        return f'<{self.__class__.__name__} client={self.client!r}, message={self.message!r}, command={self.command!r}>'
    
    def __eq__(self, other):
        """Returns whether the two contexts are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.command != other.command:
            return False
        
        if self.message is not other.message:
            return False
        
        if self.client is not other.client:
            return False
        
        return True
    
    def __hash__(self):
        """Returns the hash value of the context."""
        return hash(self.client) ^ hash(self.message) ^ hash(self.command)
    
    # Properties
    
    @property
    def channel(self):
        """
        Returns the message's channel.
        
        Returns
        -------
        channel : ``ChannelBase``
        """
        return self.message.channel
    
    
    @property
    def guild(self):
        """
        Returns the message's guild.
        
        Returns
        -------
        guild : `None` or ``Guild``
        """
        return self.message.guild
    
    
    @property
    def author(self):
        """
        Returns the message's author.
        
        Returns
        -------
        author : ``ClientUserBase``, ``Webhook``, ``WebhookRepr``
        """
        return self.message.author
    
    
    @property
    def voice_state(self):
        """
        Returns the context' user's voice state in the respective guild.
        
        Returns
        -------
        voice_state : `None` or ``VoiceState``
        """
        message = self.message
        guild = message.guild
        if (guild is not None):
            return guild.voice_states.get(message.author.id, None)
    
    
    @property
    def voice_client(self):
        """
        Returns the voice client in the message's guild if there is any.
        
        Returns
        -------
        voice_client : `None` or ``VoiceClient``
        """
        guild_id = self.message.guild_id
        if guild_id:
            voice_client = self.client.voice_clients.get(guild_id, None)
        else:
            voice_client = None
        
        return voice_client
    
    # API methods
    
    async def reply(self, *args, **kwargs):
        """
        Replies to the command's caller.
        
        This method is a coroutine.
        
        Parameters
        ----------
        content : `str`, ``EmbedBase``, `Any`, Optional
            The message's content if given. If given as `str` or empty string, then no content will be sent, meanwhile
            if any other non `str` or ``EmbedBase`` instance is given, then will be casted to string.
            
            If given as ``EmbedBase`` instance, then is sent as the message's embed.
            
        embed : ``EmbedBase`` instance or `list` of ``EmbedBase`` instances, Optional (Keyword only)
            The embedded content of the message.
            
            If `embed` and `content` parameters are both given as  ``EmbedBase`` instance, then `TypeError` is raised.
            
            If embeds are given as a list, then the first embed is picked up.
        file : `Any`, Optional (Keyword only)
            A file or files to send. Check ``create_file_form`` for details.
        sticker : `None`, ``Sticker``, `int`, (`list`, `set`, `tuple`) of (``Sticker``, `int`)
            Sticker or stickers to send within the message.
        components : `None`, ``ComponentBase``, (`set`, `list`) of ``ComponentBase``, Optional (Keyword only)
            Components attached to the message.
            
            > `components` do not count towards having any content in the message.
        allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions`` for details.
        reply_fail_fallback : `bool`, Optional (Keyword only)
            Whether normal message should be sent if the referenced message is deleted. Defaults to `False`.
        tts : `bool`, Optional (Keyword only)
            Whether the message is text-to-speech.
        nonce : `str`, Optional (Keyword only)
            Used for optimistic message sending. Will shop up at the message's data.
        
        Returns
        -------
        message : ``Message`` or `None`
            Returns `None` if there is nothing to send.
        
        Raises
        ------
        TypeError
            - If `embed` was given as `list`, but it contains not only ``EmbedBase`` instances.
            - If `allowed_mentions` contains an element of invalid type.
            - `content` parameter was given as ``EmbedBase`` instance, meanwhile `embed` parameter was given as well.
            - If invalid file type would be sent.
            - If `sticker` was not given neither as `None`, ``Sticker``, `int`, (`list`, `tuple`, `set`) of \
                (``Sticker``, `int).
        ValueError
            - If `allowed_mentions`'s elements' type is correct, but one of their value is invalid.
            - If more than `10` files would be sent.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `tts` was not given as `bool` instance.
            - If `nonce` was not given neither as `None` nor as `str` instance.
            - If `reply_fail_fallback` was not given as `bool` instance.
        """
        return await self.client.message_create(self.message, *args, **kwargs)
        
    
    async def send(self, *args, **kwargs):
        """
        Sends a message to the channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        content : `str`, ``EmbedBase``, `Any`, Optional
            The message's content if given. If given as `str` or empty string, then no content will be sent, meanwhile
            if any other non `str` or ``EmbedBase`` instance is given, then will be casted to string.
            
            If given as ``EmbedBase`` instance, then is sent as the message's embed.
            
        embed : ``EmbedBase`` instance or `list` of ``EmbedBase`` instances, Optional (Keyword only)
            The embedded content of the message.
            
            If `embed` and `content` parameters are both given as  ``EmbedBase`` instance, then `TypeError` is raised.
            
            If embeds are given as a list, then the first embed is picked up.
        file : `Any`, Optional (Keyword only)
            A file or files to send. Check ``create_file_form`` for details.
        sticker : `None`, ``Sticker``, `int`, (`list`, `set`, `tuple`) of (``Sticker``, `int`)
            Sticker or stickers to send within the message.
        components : `None`, ``ComponentBase``, (`set`, `list`) of ``ComponentBase``, Optional (Keyword only)
            Components attached to the message.
            
            > `components` do not count towards having any content in the message.
        allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions`` for details.
        tts : `bool`, Optional (Keyword only)
            Whether the message is text-to-speech.
        nonce : `str`, Optional (Keyword only)
            Used for optimistic message sending. Will shop up at the message's data.
        
        Returns
        -------
        message : ``Message`` or `None`
            Returns `None` if there is nothing to send.
        
        Raises
        ------
        TypeError
            - If `embed` was given as `list`, but it contains not only ``EmbedBase`` instances.
            - If `allowed_mentions` contains an element of invalid type.
            - `content` parameter was given as ``EmbedBase`` instance, meanwhile `embed` parameter was given as well.
            - If invalid file type would be sent.
            - If `sticker` was not given neither as `None`, ``Sticker``, `int`, (`list`, `tuple`, `set`) of \
                (``Sticker``, `int).
        ValueError
            - If `allowed_mentions`'s elements' type is correct, but one of their value is invalid.
            - If more than `10` files would be sent.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `tts` was not given as `bool` instance.
            - If `nonce` was not given neither as `None` nor as `str` instance.
            - If `reply_fail_fallback` was not given as `bool` instance.
        """
        return await self.client.message_create(self.message.channel, *args, **kwargs)
    
    
    async def typing(self):
        """
        Triggers typing indicator in the channel.
        
        This method is a coroutine.
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``ChannelTextBase`` nor `int` instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        The client will be shown up as typing for 8 seconds, or till it sends a message at the respective channel.
        """
        return await self.client.typing(self.channel)
    
    
    def keep_typing(self, *args, **kwargs):
        """
        Returns a context manager which will keep sending typing events at the channel. Can be used to indicate that
        the bot is working.
        
        Parameters
        ----------
        timeout : `float`, Optional
            The maximal duration for the ``Typer`` to keep typing.
        
        Returns
        -------
        typer : ``Typer``
        
        Examples
        --------
        ```py
        with ctx.typing():
            # Do some things
            await ctx.send('Ayaya')
        ```
        """
        return self.client.keep_typing(self.channel, *args, **kwargs)
    
    
    async def invoke(self):
        """
        Invokes the command.
        
        This method is a coroutine.
        
        Returns
        -------
        invoked : `bool`
            Whether the command was successfully invoked.
            
            If unexpected exception occurs, still returns `True`.
        """
        try:
            command_category_trace, command_function, index = get_command_category_trace(self.command, self.content, 0)
            self.command_category_trace = command_category_trace
            self.command_function = command_function
            
            if (command_function is None):
                return
            
            failed_check = await run_checks(self.command._iter_checks(), self)
            if (failed_check is not None):
                raise CommandCheckError(failed_check)
            
            for wrapper in command_function._iter_wrappers():
                await wrapper(self)
            
            parameter_parsing_states = await command_function._content_parser.parse_content(self, index)
            command_positional_parameters = self.command_positional_parameters
            command_keyword_parameters = self.command_keyword_parameters
            for parameter_parsing_state in parameter_parsing_states:
                parameter_parsing_state.get_parser_value(command_positional_parameters, command_keyword_parameters)
            
            await process_command_coroutine(self,
                command_function._function(*command_positional_parameters, **command_keyword_parameters)
            )
            
        except BaseException as err:
            return await handle_exception(self, err)
        
        return True
    
    
    @property
    def guild_id(self):
        """
        Returns the command context's message's guild's identifier.
        
        Returns
        -------
        guild_id : `int`
        """
        return self.message.guild_id
    

    @property
    def channel_id(self):
        """
        Returns the command context's message's channel's identifier.
        
        Returns
        -------
        channel_id : `int`
        """
        return self.message.channel_id



from .entity import *
from .event_types import *
from .flags import *
from .icon import *
from .preinstanced import *
from .utils import *

__all__ = (
    *entity.__all__,
    *event_types.__all__,
    *flags.__all__,
    *icon.__all__,
    *preinstanced.__all__,
    *utils.__all__,
)



"""
Checks can be added to commands or to categories to limit their usage to set users or places.


Checks can be either used as a decorator:

```py
from hata.ext.commands import checks

@Momiji.commands
@checks.owner_only()
async def knock_knock():
    return 'Awu!'
```

Or as a `.commands` parameter:

```py
from hata.ext.commands import checks

@Momiji.commands(checks=checks.owner_only())
async def knock_knock():
    return 'Awu!'
```

The implemented checks are the following:

+--------------------------------+-----------------+--------------------------------------------------------------+
| Name                           | Extra parameter | Description                                                  |
+================================+=================+==============================================================+
| announcement_channel_only      | N/A             | Whether the message's channel is an announcement channel.    |
+--------------------------------+-----------------+--------------------------------------------------------------+
| booster_only                   | N/A             | Whether the user boosts the respective guild.                |
+--------------------------------+-----------------+--------------------------------------------------------------+
| bot_account_only               | N/A             | Whether the message's author is a bot account.               |
+--------------------------------+-----------------+--------------------------------------------------------------+
| client_only                    | N/A             | Whether the message was sent by a ``Client``.                |
+--------------------------------+-----------------+--------------------------------------------------------------+
| custom                         | function        | Custom checks, to wrap a given `function`. (Can be async.)   |
+--------------------------------+-----------------+--------------------------------------------------------------+
| guild_only                     | N/A             | Whether the message was sent to a guild channel.             |
+--------------------------------+-----------------+--------------------------------------------------------------+
| guild_owner_only               | N/A             | Whether the message's author is the guild's owner.           |
|                                |                 | (Fails in private channels.)                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_any_role                   | *roles          | Whether the message's author has any of the given roles.     |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_client_guild_permissions   | permissions,    | Whether the client has the given permissions at the guild.   |
|                                | **kwargs        | (Fails in private channels.)                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_client_permissions         | permissions,    | Whether the client has the given permissions at the channel. |
|                                | **kwargs        |                                                              |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_guild_permissions          | permissions,    | Whether the message's author has the given permissions at    |
|                                | **kwargs        | the guild. (Fails in private channels.)                      |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_permissions                | permissions,    | Whether the message's author has the given permissions at    |
|                                | **kwargs        | the channel.                                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| has_role                       | role            | Whether the message's author has the given role.             |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_any_category                | *categories     | Whether the message was sent into a channel, what's category |
|                                |                 | is any of the specified ones.                                |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_any_channel                 | *channels       | Whether the message was sent to any of the given channels.   |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_any_guild                   | *guilds         | Whether the message was sent to any of the given guilds.     |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_category                    | category        | Whether the message was sent into a channel, what's category |
|                                |                 | is the specified one.                                        |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_channel                     | channel         | Whether the message's channel is the given one.              |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_guild                       | guild           | Whether the message guild is the given one.                  |
+--------------------------------+-----------------+--------------------------------------------------------------+
| is_in_voice                    | N/A             | Whether the user is in a voice channel in the respective     |
|                                |                 | guild.                                                       |
+--------------------------------+-----------------+--------------------------------------------------------------+
| release_at                     | release_at,     | Whether the command is already released. Users with the      |
|                                | *roles          | given roles and the bot owners bypass the check.             |
+--------------------------------+-----------------+--------------------------------------------------------------+
| nsfw_channel_only              | N/A             | Whether the message's channel is nsfw.                       |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_only                     | N/A             | Whether the message's author is an owner of the client.      |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_guild_owner_only      | N/A             | `owner_only` or `guild_owner` (Fails in private channels.)   |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_any_role          | *roles          | `owner_only` or `has_any_role`                               |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_guild_permissions | permissions,    | `owner_only` or `has_guild_permissions`                      |
|                                | **kwargs        | (Fails in private channels.)                                 |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_permissions       | permissions,    | `owner_only` or `has_permissions`                            |
|                                | **kwargs        |                                                              |
+--------------------------------+-----------------+--------------------------------------------------------------+
| owner_or_has_role              | role            | `owner_only` or `has_role`                                   |
+--------------------------------+-----------------+--------------------------------------------------------------+
| private_only                   | N/A             | Whether the message's channel is a private channel.          |
+--------------------------------+-----------------+--------------------------------------------------------------+
| user_account_only              | N/A             | Whether the message's author is a user account.              |
+--------------------------------+-----------------+--------------------------------------------------------------+
| user_account_or_client_only    | N/A             | Whether the message's author is a user account or a          |
|                                |                 | ``Client`` instance.                                         |
+--------------------------------+-----------------+--------------------------------------------------------------+

To handle a failed check, do:

```py
from hata.ext.commands import checks, CommandCheckError

@Momiji.commands
@checks.owner_only()
async def knock_knock():
    return 'Awu!'

@knock_knock.error
async def error_handler(ctx, exception):
    # `exception` is ``CommandCheckError`` if a check fails.
    if isinstance(err, CommandCheckError):
        await ctx.send('Owner only!')
        # Return `True` if we handled the exception and further error handles should not be called.
        return True
    
    return False
```
"""

__all__ = (
    'CheckBase',
    'announcement_channel_only',
    'booster_only',
    'bot_account_only',
    'client_only',
    'custom',
    'guild_only',
    'guild_owner_only',
    'has_any_role',
    'has_client_guild_permissions',
    'has_client_permissions',
    'has_guild_permissions',
    'has_permissions',
    'has_role',
    'is_any_category',
    'is_any_channel',
    'is_any_guild',
    'is_category',
    'is_channel',
    'is_guild',
    'is_in_voice',
    'release_at',
    'nsfw_channel_only',
    'owner_only',
    'owner_or_guild_owner_only',
    'owner_or_has_any_role',
    'owner_or_has_guild_permissions',
    'owner_or_has_permissions',
    'owner_or_has_role',
    'private_only',
    'user_account_only',
    'user_account_or_client_only',
)

from functools import partial as partial_func
from datetime import datetime


from ...backend.futures import Task
from ...backend.analyzer import CallableAnalyzer
from ...backend.utils import copy_docs
from ...backend.export import export

from ...discord.core import KOKORO
from ...discord.bases import instance_or_id_to_instance, instance_or_id_to_snowflake
from ...discord.guild import Guild
from ...discord.permission import Permission
from ...discord.role import Role
from ...discord.channel import ChannelBase, ChannelText, ChannelCategory, ChannelGuildBase, ChannelPrivate, \
    ChannelGroup
from ...discord.client import Client
from ...discord.user import ClientUserBase
from ...discord.utils import datetime_to_id

from .wrappers import CommandCheckWrapper


def _convert_permission(permission):
    """
    Validates the given `permission`.
    
    Parameters
    ----------
    permission : `None`, ``Permission`` or `int` instance.
        Permission to validate.
    
    Returns
    -------
    permissions : ``Permission``
    
    Raises
    ------
    TypeError
        If `permissions` was not given as `None`, ``Permission`` nor as `int` instance.
    """
    if permission is None:
        permission = Permission()
    else:
        if type(permission) is Permission:
            pass
        elif isinstance(permission, int):
            permission = Permission(permission)
        else:
            raise TypeError(f'`permission` can be given as `None`, `{Permission.__name__}` or as `int` instance, got '
                f'{permission.__class__.__name__}.')
    
    return permission


class CheckMeta(type):
    """
    Check metaclass for collecting their `__slots__` in a `__all_slot__` class attribute.
    """
    def __new__(cls, class_name, class_parents, class_attributes):
        """
        Parameters
        ----------
        class_name : `str`
            The created class's name.
        class_parents : `tuple` of `type` instances
            The superclasses of the creates type.
        class_attributes : `dict` of (`str`, `Any`) items
            The class attributes of the created type.
        
        Returns
        -------
        type : ``CheckMeta`` instance
        """
        if class_parents:
            parent = class_parents[0]
            inherited_slots = getattr(parent, '__all_slot__', None)
        else:
            inherited_slots = None
        
        new_slots = class_attributes.get('__slots__', None)
        
        final_slots = []
        if (inherited_slots is not None):
            final_slots.extend(inherited_slots)
        
        if (new_slots is not None):
            final_slots.extend(new_slots)
        
        class_attributes['__all_slot__'] = tuple(final_slots)
        
        return type.__new__(cls, class_name, class_parents, class_attributes)


@export
class CheckBase(metaclass=CheckMeta):
    """
    Base class for checks.
    """
    __slots__ = ()
    def __new__(cls):
        """
        Creates a new check instance.
        
        Subclasses should overwrite it.
        """
        return object.__new__(cls)
    
    async def __call__(client, context):
        """
        Returns whether the check's condition passes.
        
        Subclasses should overwrite it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        context : ``CommandContext``
            Processing context for the respective command.
        
        Returns
        -------
        passed : `bool`
            Whether the check passed.
        """
    
    def __repr__(self):
        """Returns the check's representation."""
        repr_parts = [
            self.__class__.__name__,
            '(',
        ]
        
        slots = self.__all_slot__
        limit = len(slots)
        if limit:
            index = 0
            while True:
                name = slots[index]
                index += 1
                if name.startswith('_'):
                    if index == limit:
                        break
                    else:
                        continue
                
                # case of `channel_id`, `guild_id`
                if name.endswith('id'):
                    display_name = name[:-3]
                # case of `channel_ids`, `guild_ids`
                elif name.endswith('ids'):
                    display_name = name[:-4]
                else:
                    display_name = name
                
                repr_parts.append(display_name)
                repr_parts.append('=')
                attr = getattr(self,name)
                repr_parts.append(repr(attr))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(')')
        
        return ''.join(repr_parts)
    
    def __invert__(self):
        """Inverts the check's condition returning a new check."""
        return CheckInvert(self)
    
    def __or__(self, other):
        """Connects the two check with `or` relation."""
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if type(self) is CheckBase:
            return other
        
        return CheckOrRelation(self, other)
    
    def __and__(self, other):
        """Connects the two checks which `and` relation."""
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if type(self) is CheckBase:
            return self
        
        return CheckAndRelation(self, other)


class CheckInvert(CheckBase):
    """
    Inverts the wrapped check's result.
    
    Attributes
    ----------
    check : ``CheckBase``
        The check to invert.
    """
    __slots__ = ('check', )
    def __new__(cls, check):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        check : ``CheckBase``
            The check to invert.

        Raises
        ------
        TypeError
            If `check` was not given as ``CheckBase`` instance.
        """
        if not isinstance(check, CheckBase):
            raise TypeError(f'`check` can be given as `{CheckBase.__name__}` instance, got {check.__class__.__name__}.')
        
        self = object.__new__(cls)
        self.check = check
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        return not await self.check(context)


class CheckRelationBase(CheckBase):
    """
    Base class for relation checks.
    
    Attributes
    ----------
    checks : `tuple` of ``CheckBase``
        The check to connect.
    """
    __slots__ = ('checks', )
    def __new__(cls, *checks):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *checks : ``CheckBase``
            The check to invert.

        Raises
        ------
        TypeError
            If a `check` was not given as ``CheckBase`` instance.
        """
        for check in checks:
            if not isinstance(check, CheckBase):
                raise TypeError(f'`check` can be given as `{CheckBase.__name__}` instance, got '
                    f'{check.__class__.__name__}.')
        
        self = object.__new__(cls)
        self.checks = checks
        return self


class CheckOrRelation(CheckRelationBase):
    """
    Connects checks with `or` relation.
    
    Attributes
    ----------
    checks : `tuple` of ``CheckBase``
        The check to connect.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        for check in self.checks:
            if await check(context):
                return True
        
        return False


class CheckAndRelation(CheckRelationBase):
    """
    Connects checks with `and` relation.
    
    Attributes
    ----------
    checks : `tuple` of ``CheckBase``
        The check to connect.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        for check in self.checks:
            if not await check(context):
                return False
        
        return True



class CheckSingleBase(CheckBase):
    """
    Base class for single condition checks without attributes.
    """
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if other_type is type(self):
            return self
        
        return CheckOrRelation(self, other)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if other_type is type(self):
            return self
        
        return CheckAndRelation(self, other)


class CheckIsOwner(CheckSingleBase):
    """
    Checks whether the command was called by the client's owner.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if context.client.is_owner(context.message.author):
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if other_type is type(self):
            return self
        
        # Let the other check decide
        return other|self
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if type(self) is other_type:
            return self
        
        if issubclass(other, CheckIsOwner):
            if (type(self) is CheckIsOwner):
                return self
            
            if type(other) is CheckIsOwner:
                return other
        
        return CheckAndRelation(self, other)


class CheckHasRoleBase(CheckBase):
    """
    Base class for role checks.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckHasRoleBase):
            return CheckOrRelation(self, other)
        
        roles = {*self._iter_roles(), *other._iter_roles()}
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_allowed = True
        else:
            owner_allowed = False
        
        if owner_allowed:
            check_type = CheckHasRoleOrIsOwner
        else:
            check_type = HasAnyRoleCheckOrRelationIsOwner
        
        return check_type(*roles)
    
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if not issubclass(other_type, CheckHasRoleBase):
            return CheckAndRelation(self, other)
        
        roles = {*self._iter_roles()}&{*other._iter_roles()}
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_required = True
        else:
            owner_required = False
        
        if owner_required:
            check_type = CheckHasRoleOrIsOwner
        else:
            check_type = HasAnyRoleCheckOrRelationIsOwner
        
        return check_type(*roles)
    
    
    def _iter_roles(self):
        """
        Iterates the roles of the check.
        
        This method is a generator.
        
        Yields
        ------
        role : ``Role``
        """
        return
        yield


class CheckHasRole(CheckHasRoleBase):
    """
    Checks whether the message's author has the given role.
    
    Attributes
    ----------
    role : ``Role``
        The legend itself.
    """
    __slots__ = ('role', )
    def __new__(cls, role):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        role : `str`, `int` or ``Role``
            The role what the message's author should have.
        
        Raises
        ------
        TypeError
            If `role` was not given neither as ``Role``, `str` or `int` instance.
        ValueError
            If `role` was given as `str` or as `int` instance, but not as a valid snowflake, so ``Role`` instance
            cannot be precreated with it.
        """
        role = instance_or_id_to_instance(role, Role, 'role')
        
        self = object.__new__(cls)
        self.role = role
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if context.message.author.has_role(self.role):
            return True
        
        return False
    
    @copy_docs(CheckHasRoleBase._iter_roles)
    def _iter_roles(self):
        yield self.role

class CheckHasRoleOrIsOwner(CheckHasRole, CheckIsOwner):
    """
    Checks whether the message's author has the given role, or if it the client's owner.
    
    Attributes
    ----------
    role : ``Role``
        The legend itself.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        user = context.message.author
        if user.has_role(self.role):
            return True
        
        if context.client.is_owner(user):
            return True
        
        return False


class CheckHasAnyRole(CheckHasRoleBase):
    """
    Checks whether the message's author has any of the given roles.
    
    Attributes
    ----------
    roles : `set` of ``Role``
        The legends themselves.
    """
    __slots__ = ('roles', )
    
    def __new__(cls, *roles):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *roles : `str`, `int` or ``Role``
            Role from which the message's author should have at least one.
        
        Raises
        ------
        TypeError
            If a role was not given neither as ``Role``, `str` or `int` instance.
        ValueError
            If a role was given as `str` or as `int` instance, but not as a valid snowflake, so a ``Role`` instance
            cannot be precreated with it.
        """
        roles_processed = set()
        for role in roles:
            role = instance_or_id_to_instance(role, Role, 'role')
            roles_processed.add(role)
        
        roles_processed_length = len(roles_processed)
        if roles_processed_length == 0:
            return CheckBase()
        
        if roles_processed_length == 1:
            return CheckHasRole(roles_processed.pop())
        
        self = object.__new__(cls)
        self.roles = roles_processed
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        user = context.message.author
        for role in self.roles:
            if  user.has_role(role):
                return True
        
        return False
    
    @copy_docs(CheckHasRoleBase._iter_roles)
    def _iter_roles(self):
        yield from self.roles


class HasAnyRoleCheckOrRelationIsOwner(CheckHasAnyRole, CheckIsOwner):
    """
    Checks whether the message's author has any of the given roles, or whether is it the client's owner.
    
    Attributes
    ----------
    roles : `set` of ``Role``
        The roles from what the user should have at least 1.
    """
    __slots__ = ()
    
    def __new__(cls, *roles):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *roles : `str`, `int` or ``Role``
            Role from which the message's author should have at least one.
        
        Raises
        ------
        TypeError
            If an element of role was not given neither as ``Role``, `str` or `int` instance.
        ValueError
            If a role was given as `str` or as `int` instance, but not as a valid snowflake, so a ``Role`` instance
            cannot be precreated with it.
        """
        roles_processed = set()
        for role in roles:
            role = instance_or_id_to_instance(role, Role, 'role')
            roles_processed.add(role)
        
        roles_processed_length = len(roles_processed)
        if roles_processed_length == 0:
            return CheckIsOwner()
        
        if roles_processed_length == 1:
            return CheckHasRoleOrIsOwner(roles_processed.pop())
        
        self = object.__new__(cls)
        self.roles = roles_processed
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        user = context.message.author
        for role in self.roles:
            if  user.has_role(role):
                return True
        
        if context.client.is_owner(user):
            return True
        
        return False


class CheckIsInGuild(CheckSingleBase):
    """
    Checks whether the command was called from a guild.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if isinstance(context.message.channel, ChannelGuildBase):
            return True
        
        return False
    
    @copy_docs(CheckBase.__invert__)
    def __invert__(self):
        return CheckIsInPrivate()


class CheckIsInPrivate(CheckSingleBase):
    """
    Checks whether the command was used inside of a private channel.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if isinstance(context.message.channel, (ChannelPrivate, ChannelGroup)):
            return True
        
        return False
    
    @copy_docs(CheckBase.__invert__)
    def __invert__(self):
        return CheckIsInGuild()


class CheckIsGuildOwner(CheckBase):
    """
    Checks whether the command was called by the local guild's owner.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.context
        guild = message.guild
        if guild is None:
            return False
        
        if guild.owner_id == message.author.id:
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if other_type is type(self):
            return self
        
        if other_type is not CheckIsOwner:
            return CheckOrRelation(self, other)
        
        if isinstance(self, CheckIsOwner):
            return self
        
        return CheckIsGuildOwnerOrIsOwner()


class CheckIsGuildOwnerOrIsOwner(CheckIsGuildOwner, CheckIsOwner):
    """
    Checks whether a message was sent by the message's guild's owner or by the client's owner.
    
    > Guild check is always applied.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.context
        guild = message.guild
        if guild is None:
            return False
        
        user = message.author
        if guild.owner_id == user.id:
            return True
        
        if context.client.is_owner(user):
            return True
        
        return False


class CheckHasPermissionBase(CheckBase):
    """
    Base class for checking permissions.
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    def __new__(cls, permission=None, **kwargs):
        """
        Creates a check, which will validate whether the a received message of a client passes the given condition.
        
        Parameters
        ----------
        permissions : `None`, ``Permission`` or `int` instance, Optional
            The permission, which the respective user should have. Defaults to `None`
        **kwargs : Keyword parameters
            `permission-name` - `bool` relations.
        
        Raises
        ------
        LookupError
            If a keyword is not a valid permission name.
        TypeError
            If `permission` was not given neither as `None`, ``Permission`` nor as `int` instance.
        """
        permission = _convert_permission(permission)
        permission = permission.update_by_keys(**kwargs)
        
        if not permission:
            if issubclass(cls, CheckIsOwner):
                permission_type = CheckIsOwner
            else:
                permission_type = CheckBase
            
            return permission_type()
        
        self = object.__new__(cls)
        self.permission = permission
        return self


class CheckHasPermission(CheckHasPermissionBase):
    """
    Checks whether the message's author has the given permissions in the message's channel.
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        if message.channel.permissions_for(message.author) >= self.permission:
            return True
        
        return False

    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckHasPermission):
            return CheckAndRelation(self, other)
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_allowed = True
        else:
            owner_allowed = False
        
        permission = Permission(self.permission&other.permission)
        
        if owner_allowed:
            check_type = CheckHasPermissionOrIsOwner
        else:
            check_type = CheckHasPermission
        
        return check_type(permission)
    
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckHasPermission):
            return CheckOrRelation(self, other)
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_required = True
        else:
            owner_required = False
        
        permission = Permission(self.permission|other.permission)
        if permission:
            if owner_required:
                check_type = CheckHasPermissionOrIsOwner
            else:
                check_type = CheckHasPermission
            
            return check_type(permission)
        else:
            if owner_required:
                check_type = CheckIsOwner
            else:
                check_type = CheckBase
            
            return check_type()



class CheckHasPermissionOrIsOwner(CheckHasPermission, CheckIsOwner):
    """
    Checks whether the message's author is the client's owner or has the given permissions in the message's channel.
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        user = message.author
        if message.channel.permissions_for(user) >= self.permission:
            return True
        
        if context.client.is_owner(user):
            return True
        
        return False


class CheckHasGuildPermission(CheckHasPermissionBase):
    """
    Checks whether the message's author has the given permissions in the message's guild.
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        guild = message.guild
        if guild is None:
            return False
        
        if guild.permissions_for(message.author) >= self.permission:
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckHasGuildPermission):
            return CheckOrRelation(self, other)
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_allowed = True
        else:
            owner_allowed = False
        
        permission = Permission(self.permission&other.permission)
        
        if owner_allowed:
            check_type = CheckHasGuildPermissionOrIsOwner
        else:
            check_type = CheckHasGuildPermission
        
        return check_type(permission)
    
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckHasGuildPermission):
            return CheckAndRelation(self, other)
        
        if isinstance(self, CheckIsOwner) or issubclass(other_type, CheckIsOwner):
            owner_required = True
        else:
            owner_required = False
        
        permission = Permission(self.permission|other.permission)
        if permission:
            if owner_required:
                check_type = CheckHasGuildPermissionOrIsOwner
            else:
                check_type = CheckHasGuildPermission
            
            return check_type(permission)
        else:
            if owner_required:
                check_type = CheckIsOwner
            else:
                check_type = CheckBase
            
            return check_type()


class CheckHasGuildPermissionOrIsOwner(CheckHasGuildPermission, CheckIsOwner):
    """
    Checks whether the message's author has the given permissions in the message's guild.
    
    > Guild check is always applied.
    
    Attributes
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        guild = message.guild
        if guild is None:
            return False
        
        user = message.author
        if guild.permissions_for(user) >= self.permission:
            return True
        
        if context.client.is_owner(user):
            return True
        
        return False


class CheckHasClientPermission(CheckHasPermissionBase):
    """
    Checks whether the client has the given permissions in the message's channel.
    
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if context.message.channel.cached_permissions_for(context.client) >= self.permission:
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if type(self) is not other_type:
            return CheckOrRelation(self, other)
        
        permission = Permission(self.permission&other.permission)
        
        return type(self)(permission)
    
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if type(self) is not other_type:
            return CheckOrRelation(self, other)
        
        permission = Permission(self.permission|other.permission)
        
        if not permission:
            return CheckBase()
        
        return type(self)(permission)


class CheckHasClientGuildPermission(CheckHasClientPermission):
    """
    Checks whether the client has the given permissions in the message's guild.
    
    ----------
    permission : ``Permission``
        The required permissions to pass the check.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        guild = context.message.guild
        if guild is None:
            return False
        
        if guild.cached_permissions_for(context.client) >= self.permission:
            return True
        
        return False


class CheckIsGuildBase(CheckBase):
    """
    Base class fro guild checks.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckIsGuildBase):
            return CheckOrRelation(self, other)
        
        guild_ids = {*self._iter_guild_ids(), *other._iter_guild_ids()}
        
        return CheckIsAnyGuild(*guild_ids)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if not issubclass(other_type, CheckIsGuildBase):
            return CheckAndRelation(self, other)
    
        guild_ids = {*self._iter_guild_ids()}&{*other._iter_guild_ids()}
        
        return CheckIsAnyGuild(*guild_ids)
    
    def _iter_guild_ids(self):
        """
        Iterates the guild ids of the check.
        
        This method is a generator.
        
        Yields
        ------
        guild_id : ``int``
        """
        return
        yield
    
    
class CheckIsGuild(CheckIsGuildBase):
    """
    Checks whether the message was sent to the given guild.
    
    Attributes
    ----------
    guild_id : `int`
        The respective guild's id.
    """
    __slots__ = ('guild_id', )
    def __new__(cls, guild):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        guild : `str`, `int` or ``Guild``
            The guild where the message should be sent.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild``, `str` or `int` instance.
        ValueError
            If `guild` was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        guild_id = instance_or_id_to_snowflake(guild, Guild, 'guild')
        
        self = object.__new__(cls)
        self.guild_id = guild_id
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        guild = context.message.guild
        if guild is None:
            return False
        
        if (guild.id == self.guild_id):
            return True
        
        return False
    
    @copy_docs(CheckIsGuildBase._iter_guild_ids)
    def _iter_guild_ids(self):
        yield self.guild_id


class CheckIsAnyGuild(CheckIsGuildBase):
    """
    Checks whether the message was sent into any of the given guilds.
    
    Attributes
    ----------
    guild_ids : `set` of `int`
        The respective guilds' identifiers.
    """
    __slots__ = ('guild_ids', )
    def __new__(cls, *guilds):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *guilds : `str`, `int` or ``Guild``
            The guilds where the message should be sent.
        
        Raises
        ------
        TypeError
            If a guild was not given neither as ``Guild``, `str` or `int` instance.
        ValueError
            If a guild was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        guild_ids_processed = set()
        
        for guild in guilds:
            guild_id = instance_or_id_to_snowflake(guild, Guild, 'guild')
            guild_ids_processed.add(guild_id)
        
        guild_ids_processed_length = len(guild_ids_processed)
        if guild_ids_processed_length == 0:
            return CheckBase()
        
        if guild_ids_processed_length == 1:
            return CheckIsGuild(guild_ids_processed.pop())
        
        self = object.__new__(cls)
        self.guild_ids = guild_ids_processed
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        guild = context.message.guild
        if guild is None:
            return False
        
        if (guild.id in self.guild_ids):
            return True
        
        return False
    
    @copy_docs(CheckIsGuildBase._iter_guild_ids)
    def _iter_guild_ids(self):
        yield from self.guild_ids


class CheckCustom(CheckBase):
    """
    Does a custom check.
    
    Attributes
    ----------
    _is_async : `bool`
        Whether ``.check`` is async.
    check : `callable`
        The custom check's function.
    """
    __slots__ = ('_is_async', 'check', )

    def __new__(cls, check):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        check : `callable`
            The custom check which should pass. Can be async.
        
        Raises
        ------
        TypeError
            - If `check` was not given as an `callable`.
            - If `check` accepts more or less non reserved positional non default parameters.
        
        Notes
        -----
        Only `int` instances are evaluated to boolean.
        """
        analyzer = CallableAnalyzer(check)
        non_reserved_positional_parameter_count = analyzer.get_non_reserved_positional_parameter_count()
        if  non_reserved_positional_parameter_count != 1:
            raise TypeError(f'The passed check: {check!r} should have accept `1` non reserved, positional, '
                f'non default parameters, meanwhile it accepts `{non_reserved_positional_parameter_count}`.')
        
        is_async = analyzer.is_async()
        
        self = object.__new__(cls)
        self.check = check
        self._is_async = is_async
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        try:
            result = self.check(context)
            if self._is_async:
                result = await result
        except BaseException as err:
            client = context.client
            Task(client.events.error(client, repr(self), err), KOKORO)
            return False
        
        if result is None:
            return False
        
        if isinstance(result, int) and result:
            return True
        
        return False


class CheckIsChannelBase(CheckBase):
    """
    Base class for channel checks.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckIsChannelBase):
            return CheckOrRelation(self, other)
        
        channel_ids = {*self._iter_channel_ids(), *other._iter_channel_ids()}
        
        return CheckIsAnyChannel(*channel_ids)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if not issubclass(other_type, CheckIsChannelBase):
            return CheckAndRelation(self, other)
    
        channel_ids = {*self._iter_channel_ids()}&{*other._iter_channel_ids()}
        
        return CheckIsAnyChannel(*channel_ids)
    
    def _iter_channel_ids(self):
        """
        Iterates the channel id-s of the check.
        
        This method is a generator.
        
        Yields
        ------
        channel_id : ``int``
        """
        return
        yield


class CheckIsChannel(CheckIsChannelBase):
    """
    Checks whether the message was sent to the given channel.
    
    Attributes
    ----------
    channel_id : `int`
        The respective channel's id.
    """
    __slots__ = ('channel_id', )
    def __new__(cls, channel):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        channel : `str`, `int` or ``Guild``
            The guild where the message should be sent.
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``ChannelBase``, `str` or `int` instance.
        ValueError
            If `channel` was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        channel_id = instance_or_id_to_snowflake(channel, ChannelBase, 'channel')
        
        self = object.__new__(cls)
        self.channel_id = channel_id
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if (context.message.channel.id == self.channel_id):
            return True
        
        return False
    
    @copy_docs(CheckIsChannelBase._iter_channel_ids)
    def _iter_channel_ids(self):
        yield self.channel_id


class CheckIsAnyChannel(CheckIsChannelBase):
    """
    Checks whether the message was sent into any of the given channels.
    
    Attributes
    ----------
    channel_ids : `set` of `int`
        The respective channels' identifiers.
    """
    __slots__ = ('channel_ids', )
    def __new__(cls, *channels):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *channels : `str`, `int` or ``ChannelBase``
            The channel where the message should be sent.
        
        Raises
        ------
        TypeError
            If a channel was not given neither as ``ChannelBase``, `str` or `int` instance.
        ValueError
            If a channel was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        channel_ids_processed = set()
        
        for channel in channels:
            channel_id = instance_or_id_to_snowflake(channel, ChannelBase, 'guild')
            channel_ids_processed.add(channel_id)
        
        channel_ids_processed_length = len(channel_ids_processed)
        if channel_ids_processed_length == 0:
            return CheckBase()
        
        if channel_ids_processed_length == 1:
            return CheckIsChannel(channel_ids_processed.pop())
        
        self = object.__new__(cls)
        self.channel_ids = channel_ids_processed
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if (context.message.channel.id in self.channel_ids):
            return True
        
        return False
    
    @copy_docs(CheckIsChannelBase._iter_channel_ids)
    def _iter_channel_ids(self):
        yield from self.channel_ids


class CheckIsNsfwChannel(CheckSingleBase):
    """
    Checks whether the message was sent to an nsfw channel.
    """
    __slots__ = ()

    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        channel = context.message.channel
        if (isinstance(channel, ChannelText) and channel.nsfw):
            return True
        
        return False


class CheckIsAnnouncementChannel(CheckSingleBase):
    """
    Checks whether the message was sent to an announcement channel.
    """
    __slots__ = ()

    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if context.message.channel.type == 5:
            return True
        
        return False


class CheckIsInVoice(CheckSingleBase):
    """
    Checks whether the message's author is in a voice channel in the respective guild.
    """
    __slots__ = ()

    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        guild = message.guild
        if guild is None:
            return False
        
        if message.author.id in guild.voice_states:
            return True
        
        return False


class CheckIsBooster(CheckSingleBase):
    """
    Checks whether the message's author boosts the respective guild.
    """
    __slots__ = ()

    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        guild = message.guild
        if guild is None:
            return
        
        try:
            profile = message.author.guild_profiles[guild.id]
        except KeyError:
            return False
        
        if profile.boosts_since is None:
            return False
        
        return True


class CheckIsClient(CheckSingleBase):
    """
    Check whether the message was sent by a client.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if isinstance(context.message.author, Client):
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if other_type is type(self):
            return self
        
        if other_type is CheckUserAccount:
            return CheckIsUserAccountOrIsClient()
        
        return CheckOrRelation(self, other)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if other_type is type(self):
            return self
        
        if (other_type is CheckIsUserAccountOrIsClient):
            return self
        
        return CheckAndRelation(self, other)


class CheckUserAccount(CheckSingleBase):
    """
    Checks whether the message was sent by an user account.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        if not context.message.author.is_bot:
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if other_type is type(self):
            return self
        
        if other_type is CheckIsClient:
            return CheckIsUserAccountOrIsClient()
        
        return CheckOrRelation(self, other)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if other_type is type(self):
            return self
        
        if (other_type is CheckIsUserAccountOrIsClient):
            return self
        
        return CheckAndRelation(self, other)


class CheckBotAccount(CheckSingleBase):
    """
    Checks whether the message was sent by a bot account.
    """
    __slots__ = ()

    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        user = context.message.author
        if user.is_bot and isinstance(user, ClientUserBase):
            return True
        
        return False


class CheckIsUserAccountOrIsClient(CheckIsClient, CheckUserAccount, CheckSingleBase):
    """
    Checks whether the message was sent by a user account or by a client.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        user = context.message.author
        if not user.is_bot:
            return True
        
        if isinstance(user, Client):
            return True
        
        return False
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if (other_type is type(self)) or (other_type is CheckIsClient) or (other_type is CheckUserAccount):
            return self
        
        return CheckOrRelation(self, other)

    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if other_type is type(self):
            return self
        
        if (other_type is CheckIsClient) or (other_type is CheckUserAccount):
            return other
        
        return CheckAndRelation(self, other)


class CheckIsCategoryBase(CheckBase):
    """
    Base class for category checks.
    """
    __slots__ = ()
    
    @copy_docs(CheckBase.__or__)
    def __or__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return self
        
        if not issubclass(other_type, CheckIsCategoryBase):
            return CheckOrRelation(self, other)
        
        category_ids = {*self._iter_category_ids(), *other._iter_category_ids()}
        
        return CheckIsAnyCategory(*category_ids)
    
    @copy_docs(CheckBase.__and__)
    def __and__(self, other):
        other_type = type(other)
        if not issubclass(other_type, CheckBase):
            return NotImplemented
        
        if other_type is CheckBase:
            return other
        
        if not issubclass(other_type, CheckIsCategoryBase):
            return CheckAndRelation(self, other)
        
        category_ids = {*self._iter_category_ids()}&{*other._iter_category_ids()}
        
        return CheckIsAnyCategory(*category_ids)
    
    def _iter_category_ids(self):
        """
        Iterates the category ids of the check.
        
        This method is a generator.
        
        Yields
        ------
        category_id : `int`
        """
        return
        yield


class CheckIsCategory(CheckIsCategoryBase):
    """
    Checks whether the message was sent into any channel of the given category.
    
    Attributes
    ----------
    category_id : `int`
        The respective category's id.
    """
    __slots__ = ('category_id', )
    def __new__(cls, category):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        category : `str`, `int`, ``ChannelCategory`` or ``Guild``
            The category, within sent messages pass the check.
            
            If you want to check whether the channel is not in a category, pass the parameter as the respective guild
            instead.
        
        Raises
        ------
        TypeError
            If `category` was not given neither as ``ChannelCategory``, ``Guild``, `str` or `int` instance.
        ValueError
            If `category` was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        category_id = instance_or_id_to_snowflake(category, (ChannelCategory, Guild), 'category')
        
        self = object.__new__(cls)
        self.category_id = category_id
        return self
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        channel = context.message.channel
        if not isinstance(channel, ChannelGuildBase):
            return False
        
        parent = channel.parent
        if parent is None:
            guild = channel.guild
            if guild is None:
                return False
            
            parent_id = guild.id
        else:
            parent_id = parent.id
        
        if parent_id == self.category_id:
            return True
        
        return False
    
    @copy_docs(CheckIsCategoryBase._iter_category_ids)
    def _iter_category_ids(self):
        yield self.category_id


class CheckIsAnyCategory(CheckIsCategoryBase):
    """
    Checks whether the message was sent into any channel of the given categories.
    
    Attributes
    ----------
    category_ids : `set` of `int`
        The respective categories' id.
    """
    __slots__ = ('category_ids', )
    
    def __new__(cls, *categories):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        *categories : `str`, `int`, ``ChannelCategory`` or ``Guild``
            The categories, within sent messages pass the check.
            
            If you want to check whether the channel is not in a category, pass the parameter as the respective guild
            instead.
        
        Raises
        ------
        TypeError
            If a category was not given neither as ``ChannelCategory``, ``Guild``, `str` or `int` instance.
        ValueError
            If a category was given as `str` or as `int` instance, but not as a valid snowflake.
        """
        category_ids_processed = set()
        
        for category in categories:
            category_id = instance_or_id_to_snowflake(category, (ChannelCategory, Guild), 'category')
            category_ids_processed.add(category_id)
        
        category_ids_processed_length = len(category_ids_processed)
        if category_ids_processed_length == 0:
            return CheckBase()
        if category_ids_processed_length == 1:
            return CheckIsCategory(category_ids_processed.pop())
        
        self = object.__new__(cls)
        self.category_ids = category_ids_processed
        return self

    @copy_docs(CheckIsCategoryBase._iter_category_ids)
    def _iter_category_ids(self):
        yield from self.category_ids


class CheckReleaseAt(CheckBase):
    """
    Checks whether the command is already released.
    
    Attributes
    ----------
    release_at : `int`
        The time in snowflake, when the command will be released.
    pre_access_roles : `None` or `set` of ``Role``
        The roles, who are bypassed by the check.
    """
    def __new__(cls, release_at, *roles):
        """
        Checks whether a respective condition passes.
        
        Parameters
        ----------
        release_at : `datetime`
            When the command is released.
        *roles : `str`, `int` or ``Role``
            Role from which the message's author should have at least one.
        
        Raises
        ------
        TypeError
            If an element of role was not given neither as ``Role``, `str` or `int` instance.
        ValueError
            If a role was given as `str` or as `int` instance, but not as a valid snowflake, so a ``Role`` instance
            cannot be precreated with it.
        """
        if not isinstance(release_at, datetime):
            raise TypeError(f'`release_at` can be given as `datetime` instance, got {release_at.__class__.__name__}.')
        
        roles_processed = set()
        for role in roles:
            role = instance_or_id_to_instance(role, Role, 'role')
            roles_processed.add(role)
        
        release_at = datetime_to_id(release_at)
        
        self = object.__new__(cls)
        self.release_at = release_at
        self.pre_access_roles = roles_processed
        return self
    
    
    @copy_docs(CheckBase.__call__)
    async def __call__(self, context):
        message = context.message
        if message.id > self.release_at:
            return True
        
        user = message.author
        for role in self.roles:
            if  user.has_role(role):
                return True
        
        if context.client.is_owner(user):
            return True
        
        return False


has_role = partial_func(CommandCheckWrapper, CheckHasRole)
owner_or_has_role = partial_func(CommandCheckWrapper, CheckHasRoleOrIsOwner)
has_any_role = partial_func(CommandCheckWrapper, CheckHasAnyRole)
owner_or_has_any_role = partial_func(CommandCheckWrapper, HasAnyRoleCheckOrRelationIsOwner)
guild_only = partial_func(CommandCheckWrapper, CheckIsInGuild)
private_only = partial_func(CommandCheckWrapper, CheckIsInPrivate)
owner_only = partial_func(CommandCheckWrapper, CheckIsOwner)
guild_owner_only = partial_func(CommandCheckWrapper, CheckIsGuildOwner)
owner_or_guild_owner_only = partial_func(CommandCheckWrapper, CheckIsGuildOwnerOrIsOwner)
has_permissions = partial_func(CommandCheckWrapper, CheckHasPermission)
owner_or_has_permissions = partial_func(CommandCheckWrapper, CheckHasPermissionOrIsOwner)
has_guild_permissions = partial_func(CommandCheckWrapper, CheckHasGuildPermission)
owner_or_has_guild_permissions = partial_func(CommandCheckWrapper, CheckHasGuildPermissionOrIsOwner)
has_client_permissions = partial_func(CommandCheckWrapper, CheckHasClientPermission)
has_client_guild_permissions = partial_func(CommandCheckWrapper, CheckHasClientGuildPermission)
is_guild = partial_func(CommandCheckWrapper, CheckIsGuild)
is_any_guild = partial_func(CommandCheckWrapper, CheckIsAnyGuild)
custom = partial_func(CommandCheckWrapper, CheckCustom)
is_channel = partial_func(CommandCheckWrapper, CheckIsChannel)
is_any_channel = partial_func(CommandCheckWrapper, CheckIsAnyChannel)
nsfw_channel_only = partial_func(CommandCheckWrapper, CheckIsNsfwChannel)
announcement_channel_only = partial_func(CommandCheckWrapper, CheckIsAnnouncementChannel)
is_in_voice = partial_func(CommandCheckWrapper, CheckIsInVoice)
booster_only = partial_func(CommandCheckWrapper, CheckIsBooster)
client_only = partial_func(CommandCheckWrapper, CheckIsClient)
user_account_only = partial_func(CommandCheckWrapper, CheckUserAccount)
bot_account_only = partial_func(CommandCheckWrapper, CheckBotAccount)
user_account_or_client_only = partial_func(CommandCheckWrapper, CheckIsUserAccountOrIsClient)
is_category = partial_func(CommandCheckWrapper, CheckIsCategory)
is_any_category = partial_func(CommandCheckWrapper, CheckIsAnyCategory)
release_at = partial_func(CommandCheckWrapper, CheckReleaseAt)



from .analyzer import *
from .connector import *
from .cookiejar import *
from .event_loop import *
from .exceptions import *
from .executor import *
from .export import *
from .formdata import *
from .futures import *
from .headers import *
from .helpers import *
from .http import *
from .ios import *
from .multipart import *
from .protocol import *
from .quote import *
from .reqrep import *
from .subprocess import *
from .transprotos import *
from .url import *
from .utils import *
from .websocket import *

__all__ = (
    *analyzer.__all__,
    *connector.__all__,
    *cookiejar.__all__,
    *event_loop.__all__,
    *exceptions.__all__,
    *executor.__all__,
    *export.__all__,
    *formdata.__all__,
    *futures.__all__,
    *headers.__all__,
    *helpers.__all__,
    *http.__all__,
    *ios.__all__,
    *multipart.__all__,
    *protocol.__all__,
    *quote.__all__,
    *reqrep.__all__,
    *subprocess.__all__,
    *transprotos.__all__,
    *url.__all__,
    *utils.__all__,
    *websocket.__all__,
)



__all__ = ('Closer', )

from ....backend.futures import CancelledError
from ....discord.interaction import ComponentButton, ComponentRow

from .menu import Menu
from .helpers import EMOJI_CANCEL, top_level_check, top_level_get_timeout, CUSTOM_ID_CANCEL, get_auto_check


class Closer(Menu):
    
    BUTTON_CANCEL = ComponentButton(emoji=EMOJI_CANCEL, custom_id=CUSTOM_ID_CANCEL)
    
    BUTTONS = ComponentRow(BUTTON_CANCEL,)
    
    __slots__ = ('page', 'timeout', 'user_check')
    
    def __init__(self, client, event, page, *, check=..., timeout=-1.0):
        if check is ...:
            check = get_auto_check(event)
        
        self.page = page
        self.timeout = timeout
        self.user_check = check
    
    
    check = top_level_check
    get_timeout = top_level_get_timeout
    
    async def initial_invoke(self):
        self.content = self.page
        self.components = self.BUTTONS
        self.allowed_mentions = None
    
    async def invoke(self, event):
        interaction = event.interaction
        if interaction == self.BUTTON_CANCEL:
            self.cancel(CancelledError())
        
        return False



__all__ = ()

from ..core import CHANNELS, GUILDS

def try_get_voice_region(guild_id, channel_id):
    """
    Tries to get the voice region of the given channel and guild's id-s.
    
    Parameters
    ----------
    guild_id : `int`
        The respective guild's identifier.
    channel_id : `int`
        The respective channel's identifier.
    
    Returns
    -------
    region : `None` or ``VoiceRegion``
    """
    try:
        channel = CHANNELS[channel_id]
    except KeyError:
        region = None
    else:
        region = channel.region
        if region is None:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                region = None
            else:
                region = guild.region
    
    return region



__all__ = ('ActivityRich',)

from datetime import datetime

from ..color import Color
from ..preconverters import preconvert_str, preconvert_int_options, preconvert_snowflake, preconvert_flag
from ..http import urls as module_urls
from ..utils import is_url, DISCORD_EPOCH_START, unix_time_to_datetime, datetime_to_unix_time

from . import activity_types as ACTIVITY_TYPES
from .activity_base import ActivityBase, ACTIVITY_TYPE_NAMES, ActivityAssets, ActivityParty, ActivitySecrets, \
    ActivityTimestamps, CUSTOM_IDS
from .flags import ActivityFlag

ACTIVITY_TYPE_NAME_UNKNOWN = 'unknown'

VALID_RICH_ACTIVITY_TYPES = frozenset((
    ACTIVITY_TYPES.game,
    ACTIVITY_TYPES.stream,
    ACTIVITY_TYPES.spotify,
    ACTIVITY_TYPES.watching,
    ACTIVITY_TYPES.competing,
))

class ActivityRich(ActivityBase):
    """
    Represents a Discord rich activity.
    
    Attributes
    ----------
    application_id : `int`
        The id of the activity's application. Defaults to `0`.
    assets : `None` or ``ActivityAssets``
        The activity's assets. Defaults to `None`.
    created_at : `datetime`
        When the activity was created. Defaults to Discord epoch.
    details : `None` or `str`
        What the player is currently doing. Defaults to `None`.
    flags : ``ActivityFlag``
        The flags of the activity. Defaults to `ActivityFlag(0)`
    id : `int`
        The id of the activity. Defaults to `0`.
    name : `str`
        The activity's name.
    party : `None` or ``ActivityParty``
        The activity's party.
    secrets : `None` or ``ActivitySecrets``
        The activity's secrets. Defaults to `None`.
    session_id : `None` or `str`
        Spotify activity's session's id. Defaults to `None`.
    state : `None` or `str`
        The player's current party status. Defaults to `None`.
    sync_id : `None` or `str`
        The ID of the currently playing track of a spotify activity. Defaults to `None`.
    timestamps : `None` or ``ActivityTimestamps``
        The activity's timestamps.
    type : `int`
        An integer, what represent the activity's type for Discord. Can be one of: `0`, `1`, `2`, `3`, `4`.
    url : `None` or `str`
        The url of the stream (Twitch or Youtube only). Defaults to `None`.
    """
    __slots__ = ('application_id', 'assets', 'created_at', 'details', 'flags', 'id', 'name', 'party', 'secrets',
        'session_id', 'state', 'sync_id', 'timestamps', 'type', 'url', )
    
    def __new__(cls, name, *, application_id=..., assets=None, created_at=..., details=None, flags=..., id_=...,
            party=None, secrets=None, session_id=None, state=None, sync_id=None, timestamps=None, type_=..., url=None):
        """
        Creates a new activity with the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the activity.
        application_id : `None` or `int`, Optional (Keyword only)
            The id of the activity's application. Defaults to `0`.
        assets : `None` or ``ActivityAssets``, Optional (Keyword only)
             The activity's assets. Defaults to `None`.
        details : `None` or `str`, Optional (Keyword only)
            What the player is currently doing. Defaults to `None`.
        flags : ``ActivityFlag``, `int`, Optional (Keyword only)
            The flags of the activity. Defaults to `ActivityFlag(0)`
        id_ : `int`, Optional (Keyword only)
            The id of the activity. Defaults to `0`.
        party : `None` or ``ActivityParty``, Optional (Keyword only)
            The activity's party.
        secrets : `None` or ``ActivitySecret``, Optional (Keyword only)
            The activity's secrets. Defaults to `None`.
        session_id : `None` or `str`, Optional (Keyword only)
            Spotify activity's session's id. Defaults to `None`.
        state : `None` or `str`, Optional (Keyword only)
            The player's current party status. Defaults to `None`.
        sync_id : `None` or `str`, Optional (Keyword only)
            The ID of the currently playing track of a spotify activity. Defaults to `None`.
        timestamps : `None` or ``ActivityTimestamps``, Optional (Keyword only)
            The activity's timestamps.
        type_ : `int`, Optional (Keyword only)
            The type value of the activity.
        url : `None`, `str`, Optional (Keyword only)
            The url of the activity. Only twitch and youtube urls are supported.
        
        Returns
        -------
        activity : ``ActivityRich``
        
        Raises
        ------
        TypeError
            - If `name` was not given as `str` instance.
            - If `url`'s type is neither `None` sor `str`.
            - If `type_` is not `int` instance.
            - If `application_id` is not `int` instance.
            - If `details` is neither `None` nor `str` instance.
            - If `assets` is neither `None` nor ``ActivityAssets`` instance.
            - If `party` is neither `None` nor ``ActivityParty`` instance.
            - If `id_` is not `int` instance.
            - If `created_at` is not `int` instance.
            - If `session_id` is neither `None` sor `str`
            - If `state` is neither `None` sor `str` instance.
            - if `timestamps` is neither `None` nor ``ActivityTimestamps`` instance.
            - If `sync_id` is neither `None` nor `str` instance.
        ValueError
            - If `name`'s length is out of range [1:2048].
            - If `url`'s length is out of range [0:2048].
            - If `type_` is not any of the expected values.
            - If `details`'s length is out of range [0:2048].
            - If `state`'s length is out of range [0:2048].
            - If `sync_id`'s length is out of range [0:2048].
            - If `session_id`'s length is out of range [0:2048].
        AssertionError
            - If `url` is nto a valid url.
        """
        name = preconvert_str(name, 'name', 1, 2048)
        
        if application_id is ...:
            application_id = 0
        else:
            application_id = preconvert_snowflake(application_id, 'application_id')
        
        
        if (assets is not None) and (not isinstance(assets, ActivityAssets)):
            raise TypeError(f'`assets` can be either `None` or `{ActivityAssets.__name__}` instance, got '
                f'{assets.__class__.__name__}')
        
        
        if created_at is ...:
            created_at = DISCORD_EPOCH_START
        else:
            if not isinstance(created_at, datetime):
                raise TypeError(f'`created_at` can be `datetime` instance, got {created_at.__class__.__name__}.')
        
        
        if (details is not None):
            details = preconvert_str(details, 'details', 0, 2048)
            if (not details):
                details = None
        
        
        if (flags is ...):
            flags = ActivityFlag(0)
        else:
            flags = preconvert_flag(flags, 'flags', ActivityFlag)
        
        
        
        if id_ is ...:
            id_ = 0
        else:
            id_ = preconvert_snowflake(id_, 'id_')
        
        
        if (party is not None) and (not isinstance(party, ActivityParty)):
            raise TypeError(f'`party` can be either `None` or `{ActivityParty.__name__}` instance, got '
                f'{party.__class__.__name__}.')
        
        
        if (secrets is not None) and (not isinstance(secrets, ActivitySecrets)):
            raise TypeError(f'`secrets` can be either `None` or `{ActivitySecrets.__name__}` instance, got '
                f'{secrets.__class__.__name__}.')
        
        
        if (session_id is not None):
            session_id = preconvert_str(session_id, 'session_id', 0, 2048)
            if (not session_id):
                session_id = None
        
        
        if (state is not None):
            state = preconvert_str(state, 'state', 0, 2048)
            if (not state):
                state = None
        
        
        if (timestamps is not None) and (not isinstance(timestamps, ActivityTimestamps)):
            raise TypeError(f'`timestamps` can be either `None` or `{ActivityTimestamps.__name__}` instance, got '
                f'{timestamps.__class__.__name__}')
        
        
        if (sync_id is not None):
            sync_id = preconvert_str(sync_id, 'sync_id', 0, 2048)
            if (not sync_id):
                sync_id = None
        
        
        if (url is not None):
            url = preconvert_str(url, 'url', 0, 2048)
            if url:
                if __debug__:
                    if not is_url(url):
                        raise AssertionError(f'`url` was not given as a valid url, got {url!r}.')
            
            else:
                url = None
        
        if type_ is ...:
            type_ = ACTIVITY_TYPES.game
        else:
            type_ = preconvert_int_options(type_, 'type_', VALID_RICH_ACTIVITY_TYPES)
        
        
        
        self = object.__new__(cls)
        self.name = name
        
        self.application_id = application_id
        self.details = details
        self.flags = flags
        self.state = state
        self.party = party
        self.assets = assets
        self.secrets = secrets
        self.sync_id = sync_id
        self.session_id = session_id
        self.created_at = created_at
        self.id = id_
        self.timestamps = timestamps
        self.type = type_
        self.url = url
        
        return self
    
    def __repr__(self):
        """Returns the rich activity's representation."""
        repr_parts = ['<', self.__class__.__name__, ' name=', repr(self.name), ' type=']
        
        type_value = self.type
        type_name = ACTIVITY_TYPE_NAMES.get(type_value, ACTIVITY_TYPE_NAME_UNKNOWN)
        repr_parts.append(type_name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_value))
        repr_parts.append(')')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    @property
    def color(self):
        """
        Returns the activity's color.
        
        Returns
        -------
        color : ``Color``
        """
        type_ = self.type
        if type_ == ACTIVITY_TYPES.game:
            return Color(0x7289da)
        
        if type_ == ACTIVITY_TYPES.stream:
            if self.url is None:
                return Color(0x7289da)
            else:
                return Color(0x593695)
        
        if type_ == ACTIVITY_TYPES.spotify:
            return Color(0x1db954)
        
        if type_ == ACTIVITY_TYPES.watching:
            return Color(0x7289da)
        
        return Color()
    
    @property
    def start(self):
        """
        Returns when the activity was started if applicable.
        
        Returns
        -------
        start : `None` or `datetime`
        """
        timestamps = self.timestamps
        if (timestamps is not None):
            return timestamps.start
    
    
    @property
    def end(self):
        """
        Returns when the activity ended or will end if applicable.
        
        Returns
        -------
        start : `None` or `datetime`
        """
        timestamps = self.timestamps
        if (timestamps is not None):
            return timestamps.end
    
    
    image_large_url = property(module_urls.activity_asset_image_large_url)
    image_large_url_as = module_urls.activity_asset_image_large_url_as
    image_small_url = property(module_urls.activity_asset_image_small_url)
    image_small_url_as = module_urls.activity_asset_image_small_url_as
    
    @classmethod
    def from_data(cls, activity_data):
        """
        Creates a new ``ActivityRich`` instance from the given activity data.
        
        Parameters
        ----------
        activity_data : `dict` of (`str`, `Any`) items
            Received activity data.
        
        Returns
        -------
        self : ``ActivityRich``
        """
        self = object.__new__(cls)
        type_ = activity_data['type']
        self.type = type_
        
        if type_ in CUSTOM_IDS:
            id_ = 0
        else:
            try:
                raw_id = activity_data['id']
            except KeyError:
                id_ = 0
            else:
                id_ = int(raw_id, base=16)
        
        self.id = id_
        
        self._update_attributes(activity_data)
        return self
    
    
    def _update_attributes(self, activity_data):
        """
        Updates the activity by overwriting it's old attributes.
        
        Parameters
        ----------
        activity_data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        """
        try:
            application_id = activity_data['application_id']
        except KeyError:
            application_id = 0
        else:
            application_id = int(application_id)
        self.application_id = application_id
        
        try:
            assets_data = activity_data['assets']
        except KeyError:
            assets = None
        else:
            assets = ActivityAssets.from_data(assets_data)
        self.assets = assets
        
        created_at = activity_data.get('created_at', None)
        if created_at is None:
            created_at = DISCORD_EPOCH_START
        else:
            created_at = unix_time_to_datetime(created_at)
        self.created_at = created_at
        
        self.details = activity_data.get('details', None)
        
        self.flags = ActivityFlag(activity_data.get('flags', 0))
        
        self.name = activity_data['name']
        
        try:
            party_data = activity_data['party']
        except KeyError:
            party = None
        else:
            party = ActivityParty.from_data(party_data)
        self.party = party
        
        try:
            secrets_data = activity_data['secrets']
        except KeyError:
            secrets = None
        else:
            secrets = ActivitySecrets.from_data(secrets_data)
        self.secrets = secrets
        
        self.session_id = activity_data.get('session_id', None)
        
        self.state = activity_data.get('state', None)
        
        self.sync_id = activity_data.get('sync_id', None)
        
        try:
            timestamps_data = activity_data['timestamps']
        except KeyError:
            timestamps = None
        else:
            timestamps = ActivityTimestamps.from_data(timestamps_data)
        self.timestamps = timestamps
        
        self.url = activity_data.get('url', None)
    
    
    def _difference_update_attributes(self, activity_data):
        """
        Updates the activity and returns the changes in a `dict` of (`attribute-name`, `old-value`) items.
        
        Parameters
        ----------
        activity_data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        +-------------------+-----------------------------------+
        | Keys              | Values                            |
        +===================+===================================+
        | application_id    | `int`                             |
        +-------------------+-----------------------------------+
        | assets            | `None` or ``ActivityAssets``      |
        +-------------------+-----------------------------------+
        | created_at        | `datetime`                        |
        +-------------------+-----------------------------------+
        | details           | `None` or `str`                   |
        +-------------------+-----------------------------------+
        | flags             | ``ActivityFlag``                  |
        +-------------------+-----------------------------------+
        | name              | `str`                             |
        +-------------------+-----------------------------------+
        | party             | `None` or ``ActivityParty``       |
        +-------------------+-----------------------------------+
        | secrets           | `None` or ``ActivitySecrets``     |
        +-------------------+-----------------------------------+
        | session_id        | `None` or `str`                   |
        +-------------------+-----------------------------------+
        | state             | `None` or `str`                   |
        +-------------------+-----------------------------------+
        | sync_id           | `None` or `str`                   |
        +-------------------+-----------------------------------+
        | timestamps        | `None` or `ActivityTimestamps``   |
        +-------------------+-----------------------------------+
        | url               | `None` or `str`                   |
        +-------------------+-----------------------------------+
        """
        old_attributes = {}
        
        try:
            application_id = activity_data['application_id']
        except KeyError:
            application_id = 0
        else:
            application_id = int(application_id)
        
        if self.application_id != application_id:
            old_attributes['application_id'] = self.application_id
            self.application_id = application_id
        
        try:
            assets_data = activity_data['assets']
        except KeyError:
            assets = None
        else:
            assets = ActivityAssets.from_data(assets_data)
        
        if self.assets != assets:
            old_attributes['assets'] = self.assets
            self.assets = assets
        
        created_at = activity_data.get('created_at', None)
        if created_at is None:
            created_at = DISCORD_EPOCH_START
        else:
            created_at = unix_time_to_datetime(created_at)
        if self.created_at != created_at:
            old_attributes['created_at'] = self.created_at
            self.created_at = created_at
        
        details = activity_data.get('details', None)
        if self.details != details:
            old_attributes['details'] = self.details
            self.details = details
        
        flags = activity_data.get('flags', 0)
        if self.flags != flags:
            old_attributes['flags'] = self.flags
            self.flags = ActivityFlag(flags)
        
        name = activity_data['name']
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        try:
            party_data = activity_data['party']
        except KeyError:
            party = None
        else:
            party = ActivityParty.from_data(party_data)
        
        if self.party != party:
            old_attributes['party'] = self.party
            self.party = party
        
        try:
            secrets_data = activity_data['secrets']
        except KeyError:
            secrets = None
        else:
            secrets = ActivitySecrets.from_data(secrets_data)
        
        if self.secrets != secrets:
            old_attributes['secrets'] = self.secrets
            self.secrets = secrets
        
        session_id = activity_data.get('session_id', None)
        if self.session_id != session_id:
            old_attributes['session_id'] = self.session_id
            self.session_id = session_id
        
        state = activity_data.get('state', None)
        if self.state != state:
            old_attributes['state'] = self.state
            self.state = state
        
        sync_id = activity_data.get('sync_id', None)
        if self.sync_id != sync_id:
            old_attributes['sync_id'] = self.sync_id
            self.sync_id = sync_id
        
        try:
            timestamps_data = activity_data['timestamps']
        except KeyError:
            timestamps = None
        else:
            timestamps = ActivityTimestamps.from_data(timestamps_data)
        
        if self.timestamps != timestamps:
            old_attributes['timestamps'] = self.timestamps
            self.timestamps = timestamps
        
        url = activity_data.get('url', None)
        if self.url != url:
            old_attributes['url'] = self.url
            self.url = url
        
        return old_attributes
    
    
    def bot_dict(self):
        """
        Converts the activity to json serializable dictionary, which can be sent with bot account to change activity.
        
        Returns
        -------
        activity_data : `dict` of (`str`, `Any`) items
        """
        activity_data = {
            'type' : self.type,
            'name' : self.name,
        }
        
        url = self.url
        if (url is not None):
            activity_data['url'] = url
        
        return activity_data
    
    
    def user_dict(self):
        """
        Converts the activity to json serializable dictionary, which can (?) be sent with user account to change
        activity.
        
        Returns
        -------
        activity_data : `dict` of (`str`, `Any`) items
        """
        activity_data = {}
        
        assets = self.assets
        if (assets is not None):
            assets_data = assets.to_data()
            if assets_data:
                activity_data['assets'] = assets_data
        
        details = self.details
        if (details is not None):
            activity_data['details'] = details
        
        party = self.party
        if (party is not None):
            party_data = party.to_data()
            if party_data:
                activity_data['party'] = party_data
        
        secrets = self.secrets
        if (secrets is not None):
            secrets_data = secrets.to_data()
            activity_data['secrets'] = secrets_data
        
        state = self.state
        if (state is not None):
            activity_data['state'] = state
        
        timestamps = self.timestamps
        if (timestamps is not None):
            timestamps_data = timestamps.to_data()
            if timestamps_data:
                activity_data['timestamps'] = timestamps_data
        
        
        return activity_data
    
    
    def full_dict(self):
        """
        Converts the whole activity to a dictionary.
        
        Returns
        -------
        activity_data : `dict` of (`str`, `Any`) items
        """
        activity_data = self.user_dict()
        activity_data.update(self.bot_dict())
        
        # spotify only
        flags = self.flags
        if flags:
            activity_data['flags'] = flags
        
        # spotify only
        session_id = self.session_id
        if (session_id is not None):
            activity_data['session_id'] = session_id
        
        # spotify only
        sync_id = self.sync_id
        if (sync_id is not None):
            activity_data['sync_id'] = sync_id
        
        # receive only?
        application_id = self.application_id
        if application_id:
            activity_data['application_id'] = application_id
        
        # receive only?
        activity_data['id'] = self.discord_side_id
        
        # receive only?
        created_at = self.created_at
        if created_at != DISCORD_EPOCH_START:
            activity_data['created_at'] = datetime_to_unix_time(created_at)
        
        return activity_data
    
    
    def __hash__(self):
        """Returns the activity's hash value."""
        id_ = self.id
        if id_:
            return id_
        
        # Spotify activity has no `.id`, but has `.session_id`
        return hash(self.session_id)
    
    
    @property
    def twitch_name(self):
        """
        If the user streams on twitch, returns it's twitch name.
        
        Only applicable for stream activities.
        
        Returns
        -------
        name : `None` or `str`
        """
        if self.type != ACTIVITY_TYPES.stream:
            return None
        
        assets = self.assets
        if assets is None:
            return None
        
        image_large = assets.image_large
        if image_large is None:
            return None
        
        if not image_large.startswith('twitch:'):
            return None
        
        return image_large[7:]
    
    
    @property
    def duration(self):
        """
        Returns the spotify activity's duration, or `None` if not applicable.
        
        Only applicable for spotify activities.
        
        Returns
        -------
        duration : `None` or `timedelta`
        """
        if self.type != ACTIVITY_TYPES.spotify:
            return None
        
        timestamps = self.timestamps
        start = timestamps.start
        if start is None:
            return None
        
        end = timestamps.end
        if end is None:
            return None
        
        return end-start
    
    
    @property
    def album_cover_url(self):
        """
        Returns the spotify activity's currently playing track's album url if applicable.
        
        Only applicable for spotify activities.
        
        Returns
        -------
        album_cover_url : `None` or `str`
        """
        if self.type != ACTIVITY_TYPES.spotify:
            return None
        
        assets = self.assets
        if assets is None:
            return None
        
        image_large = assets.image_large
        if image_large is None:
            return None
            
        return f'https://i.scdn.co/image/{image_large}'
    
    
    @property
    def track_id(self):
        """
        Returns the song's identifier.
        
        Only applicable for spotify activities.
        
        Returns
        -------
        track_id : `None` or `str`
        """
        if self.type != ACTIVITY_TYPES.spotify:
            return None
        
        return self.sync_id
    
    
    @property
    def track_url(self):
        """
        Returns url to the spotify activity's song.
        
        Only applicable for spotify activities.
        
        Returns
        -------
        url : `None` or `str`
        """
        if self.type != ACTIVITY_TYPES.spotify:
            return None
        
        return f'https://open.spotify.com/track/{self.sync_id}'



__all__ = ('ChannelGuildUndefined', )

from ...backend.utils import copy_docs
from ...backend.export import export, include

from ..core import CHANNELS
from ..permission import Permission
from ..permission.permission import PERMISSION_NONE, PERMISSION_MASK_VIEW_CHANNEL
from ..preconverters import preconvert_snowflake, preconvert_str, preconvert_int

from .channel_base import ChannelBase
from .channel_guild_base import ChannelGuildMainBase

parse_permission_overwrites = include('parse_permission_overwrites')

@export
class ChannelGuildUndefined(ChannelGuildMainBase):
    """
    Represents an undefined  ``Guild`` channel. This class is a place-holder for future classes. Expectedly for channel
    type `7` and `8`.
    
    Attributes
    ----------
    id : `int`
        Unique identifier of the channel.
    _permission_cache : `None` or `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    guild_id : `int`
        The channel's guild's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    type : `int`
        The channel's type.
    
    Class Attributes
    ----------------
    DEFAULT_TYPE : `int` = `7`
        The default type, what ``ChannelGuildUndefined`` represents.
    INTERCHANGE : `tuple` of `int` = `()`
        Defines to which channel type this channel's type can be interchanged. The channel's direct type must be of
        them.
    ORDER_GROUP : `int` = `0`
        An order group what defined which guild channel type comes after the other one.
    IGNORED_NAMES : `frozenset` or `str`
        Attribute names, which will not be set automatically, because they are set by other modules.
    REPRESENTED_TYPES : `tuple` = (`7`, `8`,)
        The type values which ``ChannelGuildUndefined`` might represent.
    
    Notes
    -----
    This type supports dynamic attributes.
    """
    __slots__ = ('type', '__dict__', )
    
    DEFAULT_TYPE = 7
    IGNORED_NAMES = frozenset(('type', 'name', 'position', 'permission_overwrites', ))
    INTERCHANGE = ()
    ORDER_GROUP = 0
    REPRESENTED_TYPES = (7, 8, )
    
    def __new__(cls, data, client, guild_id):
        """
        Creates an undefined guild channel from the channel data received from Discord. If the channel already exists
        and if it is partial, then updates it.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        client : `None` or ``Client``
            The client, who received the channel's data, if any.
        guild_id : `int`
            The channel's guild's identifier.
        """
        channel_id = int(data['id'])
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = channel_id
            CHANNELS[channel_id] = self
        else:
            if self.clients:
                return self
        
        self._permission_cache = None
        self.name = data['name']
        self.type = data['type']
        
        self._init_parent_and_position(data, guild_id)
        self.permission_overwrites = parse_permission_overwrites(data)
        
        for key in data.keys():
            if key in self.IGNORED_NAMES:
                continue
            
            setattr(self, key, data[key])
        
        return self
    
    
    @classmethod
    def _from_partial_data(cls, data, channel_id, guild_id):
        self = super(ChannelGuildUndefined, cls)._from_partial_data(data, channel_id, guild_id)
        
        for key in data.keys():
            if key in self.IGNORED_NAMES:
                continue
            
            setattr(self, key, data[key])
        
        return self
    
    
    @classmethod
    @copy_docs(ChannelBase._create_empty)
    def _create_empty(cls, channel_id, channel_type, guild_id):
        self = super(ChannelGuildUndefined, cls)._create_empty(channel_id, channel_type, guild_id)
        
        self.type = channel_type
        
        return self
    
    
    @property
    @copy_docs(ChannelBase.display_name)
    def display_name(self):
        return self.name
    
    
    @copy_docs(ChannelBase._update_attributes)
    def _update_attributes(self, data):
        self._permission_cache = None
        self._set_parent_and_position(data)
        self.permission_overwrites = parse_permission_overwrites(data)
        
        self.name = data['name']
        
        for key in data.keys():
            if key in self.IGNORED_NAMES:
                continue
            
            setattr(self, key, data[key])
    
    def _difference_update_attributes(self,data):
        """
        Updates the channel and returns it's overwritten attributes as a `dict` with a `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
            
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional, and it might contain extra items.
        
        Returned Data Structure
        -----------------------
        
        +---------------------------+---------------------------------------------------+
        | Keys                      | Values                                            |
        +===========================+===================================================+
        | parent_id                 | `int`                                             |
        +---------------------------+---------------------------------------------------+
        | name                      | `str`                                             |
        +---------------------------+---------------------------------------------------+
        | permission_overwrites     | `dict` of (`int`, ``PermissionOverwrite``) items  |
        +---------------------------+---------------------------------------------------+
        | position                  | `int`                                             |
        +---------------------------+---------------------------------------------------+
        """
        self._permission_cache = None
        old_attributes = {}
        
        name = data['name']
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        permission_overwrites = parse_permission_overwrites(data)
        if self.permission_overwrites != permission_overwrites:
            old_attributes['permission_overwrites'] = self.permission_overwrites
            self.permission_overwrites = permission_overwrites
        
        self._update_parent_and_position(data, old_attributes)

        for key in data.keys():
            if key in self.IGNORED_NAMES:
                continue
            
            old_value = getattr(self, key, ...)
            new_value = data[key]
            
            if old_value is ...:
                setattr(self, key, new_value)
            else:
                if old_value != new_value:
                    setattr(self, key, new_value)
                    old_attributes[key] = old_value
        
        return old_attributes
    
    
    @copy_docs(ChannelBase.permissions_for)
    def permissions_for(self,user):
        result = self._permissions_for(user)
        if not result&PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        return Permission(result)
    
    
    @copy_docs(ChannelBase.permissions_for_roles)
    def permissions_for_roles(self, *roles):
        result = self._permissions_for_roles(roles)
        if not result&PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # store channels do not have text and voice related permissions
        return Permission(result)
    
    @classmethod
    def precreate(cls, channel_id, **kwargs):
        """
        Precreates the channel by creating a partial one with the given parameters. When the channel is loaded
        the precrated channel will be picked up. If an already existing channel would be precreated, returns that
        instead and updates that only, if that is a partial channel.
        
        Parameters
        ----------
        channel_id : `int` or `str`
            The channel's id.
        **kwargs : keyword parameters
            Additional predefined attributes for the channel.
        
        Other Parameters
        ----------------
        name : `str`, Optional (Keyword only)
            The channel's ``.name``.
        type : `int`, Optional (Keyword only)
            The channel's ``.type``.
        
        Returns
        -------
        channel : ``ChannelThread``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        channel_id = preconvert_snowflake(channel_id, 'channel_id')
        
        if kwargs:
            processable = []
            
            try:
                name = kwargs.pop('name')
            except KeyError:
                pass
            else:
                name = preconvert_str(name, 'name', 2, 100)
                processable.append(('name', name))
            
            try:
                type_ = kwargs.pop('type')
            except KeyError:
                pass
            else:
                type_ = preconvert_int(type_, 'type', 0, 255)
                processable.append(('type', type_))
            
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs}')
        
        else:
            processable = None
        
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = cls._create_empty(channel_id, cls.DEFAULT_TYPE, 0)
            CHANNELS[channel_id] = self
            
        else:
            if not self.partial:
                return self
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self



__all__ = ('create_partial_webhook_from_id', )

from ...backend.export import export

from ..core import USERS

from .preinstanced import WebhookType
from .webhook import Webhook

@export
def create_partial_webhook_from_id(webhook_id, token, *, type_=WebhookType.bot, channel_id=0):
    """
    Creates a partial webhook from the given parameters. If the webhook with the given `webhook_id` already exists,
    then returns that instead.
    
    Parameters
    ----------
    webhook_id : `int`
        The identifier number of the webhook.
    token : `str`
        The token of the webhook.
    type_ : ``WebhookType``, Optional (Keyword only)
        The webhook's type. Defaults to `WebhookType.bot`.
    channel_id : `int`, Optional (Keyword only)
        The webhook's channel's identifier. Defaults to `0`.
    
    Returns
    -------
    webhook : ``Webhook``
    """
    try:
        webhook = USERS[webhook_id]
    except KeyError:
        webhook = Webhook._create_empty(webhook_id)
        webhook.channel_id = channel_id
        webhook.type = type_
        
        USERS[webhook_id] = webhook
    
    webhook.token = token
    return webhook



__all__ = ('WebhookRepr',)

from ...backend.export import export

from .utils import create_partial_webhook_from_id
from .webhook_base import WebhookBase
from .preinstanced import WebhookType

@export
class WebhookRepr(WebhookBase):
    """
    Represents a Discord webhook's user representation.
    
    Attributes
    ----------
    id : `int`
        The webhook representation's unique identifier number.
    name : str
        The webhook representation's username.
    discriminator : `int`
        The webhook representation's discriminator. Given to avoid overlapping names.
    avatar_hash : `int`
        The webhook representation's avatar's hash in `uint128`.
    avatar_type : ``IconType``
        The webhook representation's avatar's type.
    banner_color : `None` or ``Color``
        The user's banner color if has any.
    banner_hash : `int`
        The user's banner's hash in `uint128`.
    banner_type : ``IconType``
        The user's banner's type.
    channel_id : `int`
        The channel, where the webhook is going to send it's messages.
    type : ``WebhookType``
        The webhook's type.
    
    Notes
    -----
    Instances of the type support weakreferencing.
    """
    __slots__ = ()
    
    def __init__(self, data, webhook_id, type_, channel_id):
        """
        Creates a webhook representation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Included user data.
        webhook_id : `int`
            The respective webhook's identifier number.
        type_ : ``WebhookType``
            The respective webhook's type.
        channel_id : `int`
            The respective webhook's channel's identifier.
        """
        self.id = webhook_id
        self.discriminator = 0
        self.name = data['username']
        self._set_avatar(data)
        self._set_banner(data)
        self.type = type_
        self.channel_id = channel_id
        self.banner_color = None
    
    @property
    def webhook(self):
        """
        Creates a partial webhook from the webhook data included with the webhook representation.
        
        Returns
        -------
        webhook : ``Webhook``
        """
        return create_partial_webhook_from_id(self.id, '', self.type, self.channel_id)



__all__ = ('EXTENSIONS', )

import sys
from importlib.util import find_spec, module_from_spec, spec_from_file_location
from importlib import reload as reload_module

from ...backend.utils import HybridValueDictionary, WeakValueDictionary
from ...backend.export import include

from .snapshot import take_snapshot, calculate_snapshot_difference, revert_snapshot
from .utils import _validate_entry_or_exit, PROTECTED_NAMES, _get_path_extension_name
from .exceptions import DoNotLoadExtension

EXTENSION_LOADER = include('EXTENSION_LOADER')

EXTENSIONS = WeakValueDictionary()

EXTENSION_STATE_UNDEFINED = 0
EXTENSION_STATE_LOADED = 1
EXTENSION_STATE_UNLOADED = 2
EXTENSION_STATE_UNSATISFIED = 3

EXTENSION_STATE_VALUE_TO_NAME = {
    EXTENSION_STATE_UNDEFINED: 'undefined',
    EXTENSION_STATE_LOADED: 'loaded',
    EXTENSION_STATE_UNLOADED: 'unloaded',
    EXTENSION_STATE_UNSATISFIED: 'unsatisfied',
}

class Extension:
    """
    Represents an extension.
    
    Attributes
    ----------
    _added_variable_names : `list of `str`
        A list of the added variables' names to the module.
    _default_variables : `None` or `HybridValueDictionary` of (`str`, `Any`) items
        An optionally weak value dictionary to store objects for assigning them to modules before loading them.
        If it would be set as empty, then it is set as `None` instead.
    _entry_point : `None`, `str`, or `callable`
        Internal slot used by the ``.entry_point`` property.
    _exit_point : `None`, `str`, or `callable`
        Internal slot used by the ``.exit_point`` property.
    _extend_default_variables : `bool`
        Internal slot used by the ``.extend_default_variables`` property.
    _lib : `None` or module`
        The extension's module. Set as `module` object if it the extension was already loaded.
    _locked : `bool`
        The internal slot used for the ``.locked`` property.
    _snapshot_difference : `None` or `list` of `tuple` (``Client``, `list` of `tuple` (`str`, `Any`))
        Snapshot difference if applicable. Defaults to `None`.
    _spec : `ModuleSpec`
        The module specification for the extension's module's import system related state.
    _state : `int`
        The state of the extension. Can be:
        +-------------------------------+-------+
        | Respective name               | Value |
        +===============================+=======+
        | EXTENSION_STATE_UNDEFINED     | 0     |
        +-------------------------------+-------+
        | EXTENSION_STATE_LOADED        | 1     |
        +-------------------------------+-------+
        | EXTENSION_STATE_UNLOADED      | 2     |
        +-------------------------------+-------+
        | EXTENSION_STATE_UNSATISFIED   | 3     |
        +-------------------------------+-------+
    _take_snapshot : `bool`
        Whether snapshot difference should be taken.
    """
    __slots__ = ('__weakref__', '_added_variable_names', '_default_variables', '_entry_point', '_exit_point',
        '_extend_default_variables', '_lib', '_locked', '_snapshot_difference', '_spec', '_state',
        '_take_snapshot', )
    
    def __new__(cls, name, path, entry_point, exit_point, extend_default_variables, locked, take_snapshot_difference,
            default_variables, ):
        """
        Creates an extension with the given parameters. If an extension already exists with the given name, returns
        that.
        
        Parameters
        ----------
        name : `None` or `str`
            The extension's name (or import path).
        path : `str`
            Path to the extension file.
        entry_point : `None`, `str` or `callable`
            The entry point of the extension.
        exit_point : `None`, `str` or `callable`
            The exit point of the extension.
        extend_default_variables : `bool`
            Whether the extension should use the loader's default variables or just it's own's.
        locked : `bool`
            Whether the extension should be picked up by the `{}_all` methods of the extension loader.
        take_snapshot_difference: `bool`
            Whether snapshots should be taken before and after loading an extension, and when the extension is unloaded,
            the snapshot difference should be reverted.
        default_variables : `None` or `HybridValueDictionary` of (`str`, `Any`) items
            An optionally weak value dictionary to store objects for assigning them to modules before loading them.
            If would be empty, is set as `None` instead.
        
        Returns
        -------
        self : ``Extension``
        
        Raises
        ------
        ModuleNotFoundError
            If the extension was not found.
        """
        if (name is None):
            name = _get_path_extension_name(path)
            
            from_path = True
        else:
            from_path = False
        
        try:
            return EXTENSIONS[name]
        except KeyError:
            pass
        
        if from_path:
            spec = spec_from_file_location(name, path)
        else:
            spec = find_spec(name)
        
        if spec is None:
            raise ModuleNotFoundError(name)
        
        self = object.__new__(cls)
        self._state = EXTENSION_STATE_UNDEFINED
        self._spec = spec
        self._lib = None
        self._entry_point = entry_point
        self._exit_point = exit_point
        self._extend_default_variables = extend_default_variables
        self._locked = locked
        self._default_variables = default_variables
        self._added_variable_names = []
        self._take_snapshot = take_snapshot_difference
        self._snapshot_difference = None
        
        EXTENSIONS[name] = self
        
        return self
    
    
    def __hash__(self):
        """Returns the extension's ``._spec``'s `.origin`'s hash."""
        return hash(self._spec.origin)
    
    
    def __repr__(self):
        """Returns the extension's representation."""
        repr_parts = []
        repr_parts.append('<')
        repr_parts.append(self.__class__.__name__)
        repr_parts.append(' name=')
        repr_parts.append(repr(self._spec.name))
        
        state = self._state
        repr_parts.append(', state=')
        state_name = EXTENSION_STATE_VALUE_TO_NAME.get(state, '???')
        repr_parts.append(state_name)
        repr_parts.append(' (')
        repr_parts.append(repr(state))
        repr_parts.append(')')
        
        if self._locked:
            repr_parts.append(', locked=True')
        
        default_variables = self._default_variables
        if self._extend_default_variables:
            if (default_variables is not None):
                repr_parts.append(' extends loader\'s defaults with: ')
                repr_parts.append(repr(default_variables))
        else:
            if default_variables is None:
                repr_parts.append(' clears loader\'s defaults')
            else:
                repr_parts.append(' clears loader\'s defaults and uses: ')
                repr_parts.append(repr(default_variables))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def add_default_variables(self, **variables):
        """
        Adds default variables to the extension.
        
        Parameters
        ----------
        **variables : Keyword Parameters
            Variables to assigned to the extension's module before it is loaded.
        
        Raises
        ------
        ValueError
             If a variable name is would be used, what is `module` attribute.
        """
        if not variables:
            return
        
        default_variables = self._default_variables
        if default_variables is None:
            default_variables = HybridValueDictionary()
            self._default_variables = default_variables
        
        for key, value in variables.items():
            if key in PROTECTED_NAMES:
                raise ValueError(f'The passed {key!r} is a protected variable name of module type.')
            default_variables[key] = value
    
    
    def remove_default_variables(self, *names):
        """
        Removes the mentioned default variables of the extension.
        
        If a variable with a specified name is not found, no error is raised.
        
        Parameters
        ----------
        *names : `str`
            Default variable names.
        """
        default_variables = self._default_variables
        if default_variables is None:
            return
        
        for name in names:
            try:
                del default_variables[name]
            except KeyError:
                pass
        
        if default_variables:
            return
        
        self._default_variables = None
    
    
    def clear_default_variables(self):
        """
        Removes all the default variables of the extension.
        """
        self._default_variables = None
    
    
    @property
    def entry_point(self):
        """
        Get-set-del descriptor for modifying the extension's entry point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._entry_point
    
    
    @entry_point.setter
    def entry_point(self, entry_point):
        if not _validate_entry_or_exit(entry_point):
            raise TypeError(f'`{self.__class__.__name__}.entry_point` expected `None`, `str` or a `callable`, got '
                f'{entry_point.__class__.__name__}.')
        
        self._entry_point = entry_point
    
    @entry_point.deleter
    def entry_point(self):
        self._entry_point = None
    
    
    @property
    def exit_point(self):
        """
        Get-set-del descriptor for modifying the extension's exit point.
        
        Accepts and returns `None`, `str` or a `callable`. If invalid type is given, raises `TypeError`.
        """
        return self._exit_point
    
    @exit_point.setter
    def exit_point(self, exit_point):
        if not _validate_entry_or_exit(exit_point):
            raise TypeError(f'`{self.__class__.__name__}.exit_point` expected `None`, `str` or a `callable`, got '
                f'{exit_point.__class__.__name__}.')
        
        self._exit_point = exit_point
    
    @exit_point.deleter
    def exit_point(self):
        self._exit_point = None
    
    
    @property
    def extend_default_variables(self):
        """
        Get-set descriptor to define whether the extension uses the loader's default variables or just it's own's.
        
        Accepts and returns `bool`.
        """
        return self._extend_default_variables
    
    @extend_default_variables.setter
    def extend_default_variables(self, extend_default_variables):
        extend_default_variables_type = extend_default_variables.__class__
        if extend_default_variables_type is bool:
            pass
        elif issubclass(extend_default_variables_type, int):
            extend_default_variables = bool(extend_default_variables)
        else:
            raise TypeError(f'`extend_default_variables` should have been passed as `bool`, got '
                f'{extend_default_variables_type.__name__}.')
        
        self._extend_default_variables = extend_default_variables
    
    @property
    def locked(self):
        """
        Get-set property to define whether the extension should be picked up by the `{}_all` methods of the
        extension loader.
        
        Accepts and returns `bool`.
        """
        return self._locked
    
    @locked.setter
    def locked(self, locked):
        locked_type = locked.__class__
        if locked_type is bool:
            pass
        elif issubclass(locked_type, int):
            locked = bool(locked)
        else:
            raise TypeError(f'`locked` should have been passed as `bool`, got: {locked_type.__name__}.')
        
        self._locked = locked
    
    
    def _load(self):
        """
        Loads the module and returns it. If it is already loaded returns `None`.
        
        Returns
        -------
        lib : `None` or `lib`
        """
        state = self._state
        if state == EXTENSION_STATE_UNDEFINED:
            
            spec = self._spec
            lib = sys.modules.get(spec.name, None)
            if lib is None:
                # lib is not imported yet, nice
                lib = module_from_spec(spec)
                
                added_variable_names = self._added_variable_names
                if self._extend_default_variables:
                    for name, value in EXTENSION_LOADER._default_variables.items():
                        setattr(lib, name, value)
                        added_variable_names.append(name)
                
                default_variables = self._default_variables
                if (default_variables is not None):
                    for name, value in default_variables.items():
                        setattr(lib, name, value)
                        added_variable_names.append(name)
                
                if self._take_snapshot:
                    snapshot_old = take_snapshot()
                
                try:
                    spec.loader.exec_module(lib)
                except DoNotLoadExtension:
                    loaded = False
                else:
                    loaded = True
                
                sys.modules[spec.name] = lib
                
                if loaded:
                    if self._take_snapshot:
                        snapshot_new = take_snapshot()
                        
                        self._snapshot_difference = calculate_snapshot_difference(snapshot_old, snapshot_new)
            
            else:
                loaded = True
            
            self._lib = lib
            
            if loaded:
                state = EXTENSION_STATE_LOADED
            else:
                state = EXTENSION_STATE_UNSATISFIED
                lib = None
            
            self._state = state
            return lib
        
        if state == EXTENSION_STATE_LOADED:
            # return None -> already loaded
            return None
        
        if state in (EXTENSION_STATE_UNLOADED, EXTENSION_STATE_UNSATISFIED):
            # reload
            lib = self._lib
            
            added_variable_names = self._added_variable_names
            if self._extend_default_variables:
                for name, value in EXTENSION_LOADER._default_variables.items():
                    setattr(lib, name, value)
                    added_variable_names.append(name)
            
            default_variables = self._default_variables
            if (default_variables is not None):
                for name, value in default_variables.items():
                    setattr(lib, name, value)
                    added_variable_names.append(name)
            
            if self._take_snapshot:
                snapshot_old = take_snapshot()
            
            try:
                reload_module(lib)
            except DoNotLoadExtension:
                loaded = False
            else:
                loaded = True
            
            if loaded:
                if self._take_snapshot:
                    snapshot_new = take_snapshot()
                    
                    self._snapshot_difference = calculate_snapshot_difference(snapshot_old, snapshot_new)
            
            if loaded:
                state = EXTENSION_STATE_LOADED
            else:
                state = EXTENSION_STATE_UNSATISFIED
                lib = None
            
            self._state = state
            
            return lib
        
        return None
        # no more cases
    
    def _unload(self):
        """
        Unloads the module and returns it. If it is already unloaded returns `None`.
        
        Returns
        -------
        lib : `None` or `lib`
        """
        state = self._state
        if state == EXTENSION_STATE_UNDEFINED:
            return None
        
        if state == EXTENSION_STATE_LOADED:
            
            snapshot_difference = self._snapshot_difference
            if (snapshot_difference is not None):
                self._snapshot_difference = None
                revert_snapshot(snapshot_difference)
            
            self._state = EXTENSION_STATE_UNLOADED
            return self._lib
        
        if state in (EXTENSION_STATE_UNLOADED, EXTENSION_STATE_UNSATISFIED):
            return None
        
        # no more cases
    
    def _unassign_variables(self):
        """
        Unassigns the assigned variables to the respective module and clears ``._added_variable_names``.
        """
        lib = self._lib
        if lib is None:
            return
        
        added_variable_names = self._added_variable_names
        for name in added_variable_names:
            try:
                delattr(lib, name)
            except AttributeError:
                pass
        added_variable_names.clear()
    
    @property
    def name(self):
        """
        Returns the extension's name.
        
        Returns
        -------
        name : `str`
        """
        return self._spec.name
    
    @property
    def short_name(self):
        """
        Returns the extension's name's shortened version.
        
        Returns
        -------
        name : `str`
        """
        name = self._spec.name
        dot_index = name.rfind('.')
        if dot_index == -1:
            short_name = None
        else:
            short_name = name[dot_index+1:]
        
        return short_name
    
    
    def is_loaded(self):
        """
        Returns whether the extension is loaded.
        
        Returns
        -------
        is_loaded : `bool`
        """
        return (self._state == EXTENSION_STATE_LOADED)
    
   
    def is_unsatisfied(self):
        """
        Returns whether the extension is unsatisfied.
        
        Returns
        -------
        is_unsatisfied : `bool`
        """
        return (self._state == EXTENSION_STATE_UNSATISFIED)
    
    
    def _unlink(self):
        """
        Removes the extension's module from the loaded ones.
        
        Should not be called on loaded extensions.
        """
        state = self._state
        if state == EXTENSION_STATE_UNDEFINED:
            return
        
        if state == EXTENSION_STATE_LOADED:
            self._state = EXTENSION_STATE_UNLOADED
            lib = self._lib
            
            if self._extend_default_variables:
                for name in EXTENSION_LOADER._default_variables:
                    try:
                        delattr(lib, name)
                    except AttributeError:
                        pass
            
            default_variables = self._default_variables
            if (default_variables is not None):
                for name in default_variables:
                    delattr(lib, name)
            
            try:
                del sys.modules[self._spec.name]
            except KeyError:
                pass
            return
        
        if state in (EXTENSION_STATE_UNLOADED, EXTENSION_STATE_UNSATISFIED):
            try:
                del sys.modules[self._spec.name]
            except KeyError:
                pass
            return
        
        # no more cases



__all__ = ('IntentFlag',)

from ...env import CACHE_PRESENCE

from ..bases import FlagBase

INTENT_SHIFT_GUILDS = 0
INTENT_SHIFT_GUILD_USERS = 1
INTENT_SHIFT_GUILD_BANS = 2
INTENT_SHIFT_GUILD_EMOJIS_AND_STICKERS = 3
INTENT_SHIFT_GUILD_INTEGRATIONS = 4
INTENT_SHIFT_GUILD_WEBHOOKS = 5
INTENT_SHIFT_GUILD_INVITES = 6
INTENT_SHIFT_GUILD_VOICE_STATES = 7
INTENT_SHIFT_GUILD_PRESENCES = 8
INTENT_SHIFT_GUILD_MESSAGES = 9
INTENT_SHIFT_GUILD_REACTIONS = 10
INTENT_SHIFT_GUILD_TYPINGS = 11
INTENT_SHIFT_DIRECT_MESSAGES = 12
INTENT_SHIFT_DIRECT_REACTIONS = 13
INTENT_SHIFT_DIRECT_TYPINGS = 14

INTENT_MASK_GUILDS = 1<<INTENT_SHIFT_GUILDS
INTENT_MASK_GUILD_USERS = 1<<INTENT_SHIFT_GUILD_USERS
INTENT_MASK_GUILD_BANS = 1<<INTENT_SHIFT_GUILD_BANS
INTENT_MASK_GUILD_EMOJIS_AND_STICKERS = 1<<INTENT_SHIFT_GUILD_EMOJIS_AND_STICKERS
INTENT_MASK_GUILD_INTEGRATIONS = 1<<INTENT_SHIFT_GUILD_INTEGRATIONS
INTENT_MASK_GUILD_WEBHOOKS = 1<<INTENT_SHIFT_GUILD_WEBHOOKS
INTENT_MASK_GUILD_INVITES = 1<<INTENT_SHIFT_GUILD_INVITES
INTENT_MASK_GUILD_VOICE_STATES = 1<<INTENT_SHIFT_GUILD_VOICE_STATES
INTENT_MASK_GUILD_PRESENCES = 1<<INTENT_SHIFT_GUILD_PRESENCES
INTENT_MASK_GUILD_MESSAGES = 1<<INTENT_SHIFT_GUILD_MESSAGES
INTENT_MASK_GUILD_REACTIONS = 1<<INTENT_SHIFT_GUILD_REACTIONS
INTENT_MASK_GUILD_TYPINGS = 1<<INTENT_SHIFT_GUILD_TYPINGS
INTENT_MASK_DIRECT_MESSAGES = 1<<INTENT_SHIFT_DIRECT_MESSAGES
INTENT_MASK_DIRECT_REACTIONS = 1<<INTENT_SHIFT_DIRECT_REACTIONS
INTENT_MASK_DIRECT_TYPINGS = 1<<INTENT_SHIFT_DIRECT_TYPINGS


INTENT_SHIFT_EVENTS = {
    INTENT_SHIFT_GUILDS: (
        'GUILD_CREATE',
        'GUILD_DELETE',
        'GUILD_UPDATE',
        'GUILD_ROLE_CREATE',
        'GUILD_ROLE_UPDATE',
        'GUILD_ROLE_DELETE',
        'CHANNEL_CREATE',
        'CHANNEL_UPDATE',
        'CHANNEL_DELETE',
        'CHANNEL_PINS_UPDATE',
        'STAGE_INSTANCE_CREATE',
        'STAGE_INSTANCE_UPDATE',
        'STAGE_INSTANCE_DELETE',
        'THREAD_CREATE',
        'THREAD_UPDATE',
        'THREAD_DELETE',
        'THREAD_LIST_SYNC',
        'THREAD_MEMBER_UPDATE',
        'THREAD_MEMBERS_UPDATE',
        'GUILD_APPLICATION_COMMAND_COUNTS_UPDATE',
        'GUILD_SCHEDULED_EVENT_CREATE',
        'GUILD_SCHEDULED_EVENT_UPDATE',
        'GUILD_SCHEDULED_EVENT_DELETE',
        'GUILD_SCHEDULED_EVENT_USER_CREATE',
        'GUILD_SCHEDULED_EVENT_USER_DELETE',
    ),
    INTENT_SHIFT_GUILD_USERS: (
        'GUILD_MEMBER_ADD',
        'GUILD_MEMBER_UPDATE',
        'GUILD_MEMBER_REMOVE',
        'GUILD_JOIN_REQUEST_DELETE',
        'THREAD_MEMBERS_UPDATE',
    ),
    INTENT_SHIFT_GUILD_BANS: (
        'GUILD_BAN_ADD',
        'GUILD_BAN_REMOVE',
    ),
    INTENT_SHIFT_GUILD_EMOJIS_AND_STICKERS: (
        'GUILD_EMOJIS_UPDATE',
        'GUILD_STICKERS_UPDATE',
    ),
    INTENT_SHIFT_GUILD_INTEGRATIONS: (
        'INTEGRATION_CREATE',
        'INTEGRATION_DELETE',
        'INTEGRATION_UPDATE',
        'GUILD_INTEGRATIONS_UPDATE',
    ),
    INTENT_SHIFT_GUILD_WEBHOOKS: (
        'WEBHOOKS_UPDATE',
    ),
    INTENT_SHIFT_GUILD_INVITES: (
        'INVITE_CREATE',
        'INVITE_DELETE',
    ),
    INTENT_SHIFT_GUILD_VOICE_STATES: (
        'VOICE_STATE_UPDATE',
    ),
    INTENT_SHIFT_GUILD_PRESENCES: (
        'PRESENCE_UPDATE',
    ),
    INTENT_SHIFT_GUILD_MESSAGES: (
        'CHANNEL_PINS_UPDATE',
        'MESSAGE_CREATE',
        'MESSAGE_UPDATE',
        'MESSAGE_DELETE',
        'MESSAGE_DELETE_BULK', # Not listed by Discord, yayyyy
    ),
    INTENT_SHIFT_GUILD_REACTIONS: (
        'MESSAGE_REACTION_ADD',
        'MESSAGE_REACTION_REMOVE',
        'MESSAGE_REACTION_REMOVE_ALL',
        'MESSAGE_REACTION_REMOVE_EMOJI',
    ),
    INTENT_SHIFT_GUILD_TYPINGS: (
        'TYPING_START',
    ),
    INTENT_SHIFT_DIRECT_MESSAGES: (
        'CHANNEL_CREATE',
        'CHANNEL_PINS_UPDATE',
        'MESSAGE_CREATE',
        'MESSAGE_UPDATE',
        'MESSAGE_DELETE',
    ),
    INTENT_SHIFT_DIRECT_REACTIONS: (
        'MESSAGE_REACTION_ADD',
        'MESSAGE_REACTION_REMOVE',
        'MESSAGE_REACTION_REMOVE_ALL',
        'MESSAGE_REACTION_REMOVE_EMOJI',
    ),
    INTENT_SHIFT_DIRECT_TYPINGS : (
        'TYPING_START',
    ),
}

GLOBAL_INTENT_SHIFT_EVENTS = (
    'READY',
    'RESUMED',
    'USER_UPDATE',
    'CHANNEL_RECIPIENT_ADD', # User account only
    'CHANNEL_RECIPIENT_REMOVE', # User only
    'GUILD_MEMBERS_CHUNK',
    'VOICE_SERVER_UPDATE',
    'RELATIONSHIP_ADD', # User account only
    'RELATIONSHIP_REMOVE', # User account only
    'PRESENCES_REPLACE', # Empty / User account
    'USER_SETTINGS_UPDATE', # User account only
    'GIFT_CODE_UPDATE',
    'USER_ACHIEVEMENT_UPDATE', # User account only
    'MESSAGE_ACK', # User account only
    'SESSIONS_REPLACE', # User account only
    'INTERACTION_CREATE',
    'APPLICATION_COMMAND_CREATE',
    'APPLICATION_COMMAND_UPDATE',
    'APPLICATION_COMMAND_DELETE',
    'APPLICATION_COMMAND_PERMISSIONS_UPDATE',
    'USER_GUILD_SETTINGS_UPDATE', # User account only
    'CHANNEL_UNREAD_UPDATE', # User account only
)

INTENT_SHIFT_DEFAULT_EVENT = 255
INTENT_SHIFT_MISSING_EVENT = 254


DISPATCH_EVENT_TO_INTENTS = {}

def populate_dispatch_event_intents():
    from itertools import chain
    
    for intent_shift, event_names in chain(
        INTENT_SHIFT_EVENTS.items(), ((INTENT_SHIFT_DEFAULT_EVENT, GLOBAL_INTENT_SHIFT_EVENTS),)
    ):
        
        for event_name in event_names:
            try:
                intent_shifts = DISPATCH_EVENT_TO_INTENTS[event_name]
            except KeyError:
                intent_shifts = (intent_shift, )
            else:
                intent_shifts = (*intent_shifts, intent_shift, )
            
            DISPATCH_EVENT_TO_INTENTS[event_name] = intent_shifts

populate_dispatch_event_intents()
del populate_dispatch_event_intents


class IntentFlag(FlagBase, enable_keyword='allow', disable_keyword='deny'):
    """
    An `int` subclass representing the intents to receive specific events. The wrapper picks these up as well and
    optimizes the dispatch events' events.
    
    Each flag specifies which parser's dispatch event is received from Discord. Not mentioned events do not depend
    on intent flags and they are expected to be received independently.
    
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | Intent flag position's                 | Shift | Intent name                | Corresponding parser                        |
    | respective name                        | value |                            |                                             |
    +========================================+=======+============================+=============================================+
    | INTENT_SHIFT_GUILDS                    | 0     | guilds                     | GUILD_CREATE,                               |
    |                                        |       |                            | GUILD_DELETE,                               |
    |                                        |       |                            | GUILD_UPDATE,                               |
    |                                        |       |                            | GUILD_ROLE_CREATE,                          |
    |                                        |       |                            | GUILD_ROLE_UPDATE,                          |
    |                                        |       |                            | GUILD_ROLE_DELETE,                          |
    |                                        |       |                            | CHANNEL_CREATE,                             |
    |                                        |       |                            | CHANNEL_UPDATE,                             |
    |                                        |       |                            | CHANNEL_DELETE,                             |
    |                                        |       |                            | CHANNEL_PINS_UPDATE,                        |
    |                                        |       |                            | STAGE_INSTANCE_CREATE,                      |
    |                                        |       |                            | STAGE_INSTANCE_UPDATE,                      |
    |                                        |       |                            | STAGE_INSTANCE_DELETE,                      |
    |                                        |       |                            | THREAD_CREATE,                              |
    |                                        |       |                            | THREAD_UPDATE,                              |
    |                                        |       |                            | THREAD_DELETE,                              |
    |                                        |       |                            | THREAD_LIST_SYNC,                           |
    |                                        |       |                            | THREAD_MEMBER_UPDATE,                       |
    |                                        |       |                            | THREAD_MEMBERS_UPDATE,                      |
    |                                        |       |                            | GUILD_APPLICATION_COMMAND_COUNTS_UPDATE,    |
    |                                        |       |                            | GUILD_SCHEDULED_EVENT_CREATE,               |
    |                                        |       |                            | GUILD_SCHEDULED_EVENT_UPDATE,               |
    |                                        |       |                            | GUILD_SCHEDULED_EVENT_DELETE,               |
    |                                        |       |                            | GUILD_SCHEDULED_EVENT_USER_CREATE,          |
    |                                        |       |                            | GUILD_SCHEDULED_EVENT_USER_DELETE,          |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_USERS               | 1     | guild_users                | GUILD_MEMBER_ADD,                           |
    |                                        |       |                            | GUILD_MEMBER_UPDATE,                        |
    |                                        |       |                            | GUILD_MEMBER_REMOVE,                        |
    |                                        |       |                            | GUILD_JOIN_REQUEST_DELETE,                  |
    |                                        |       |                            | THREAD_MEMBERS_UPDATE                       |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_BANS                | 2     | guild_bans                 | GUILD_BAN_ADD,                              |
    |                                        |       |                            | GUILD_BAN_REMOVE                            |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_EMOJIS_AND_STICKERS | 3     | guild_emojis_and_stickers  | GUILD_EMOJIS_UPDATE                         |
    |                                        |       |                            | GUILD_STICKERS_UPDATE                       |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_INTEGRATIONS        | 4     | guild_integrations         | INTEGRATION_CREATE,                         |
    |                                        |       |                            | INTEGRATION_DELETE,                         |
    |                                        |       |                            | INTEGRATION_UPDATE,                         |
    |                                        |       |                            | GUILD_INTEGRATIONS_UPDATE                   |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_WEBHOOKS            | 5     | guild_webhooks             | WEBHOOKS_UPDATE                             |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_INVITES             | 6     | guild_invites              | INVITE_CREATE,                              |
    |                                        |       |                            | INVITE_DELETE                               |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_VOICE_STATES        | 7     | guild_voice_states         | VOICE_STATE_UPDATE                          |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_PRESENCES           | 8     | guild_presences            | PRESENCE_UPDATE                             |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_MESSAGES            | 9     | guild_messages             | CHANNEL_PINS_UPDATE,                        |
    |                                        |       |                            | MESSAGE_CREATE,                             |
    |                                        |       |                            | MESSAGE_UPDATE,                             |
    |                                        |       |                            | MESSAGE_DELETE,                             |
    |                                        |       |                            | MESSAGE_DELETE_BULK                         |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_REACTIONS           | 10    | guild_reactions            | MESSAGE_REACTION_ADD,                       |
    |                                        |       |                            | MESSAGE_REACTION_REMOVE,                    |
    |                                        |       |                            | MESSAGE_REACTION_REMOVE_ALL,                |
    |                                        |       |                            | MESSAGE_REACTION_REMOVE_EMOJI               |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_GUILD_TYPINGS             | 11    | guild_typings              | TYPING_START                                |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_DIRECT_MESSAGES           | 12    | direct_messages            | CHANNEL_CREATE,                             |
    |                                        |       |                            | CHANNEL_PINS_UPDATE,                        |
    |                                        |       |                            | MESSAGE_CREATE,                             |
    |                                        |       |                            | MESSAGE_UPDATE,                             |
    |                                        |       |                            | MESSAGE_DELETE                              |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_DIRECT_REACTIONS          | 13    | direct_reactions           | MESSAGE_REACTION_ADD,                       |
    |                                        |       |                            | MESSAGE_REACTION_REMOVE,                    |
    |                                        |       |                            | MESSAGE_REACTION_REMOVE_ALL,                |
    |                                        |       |                            | MESSAGE_REACTION_REMOVE_EMOJI               |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    | INTENT_SHIFT_DIRECT_TYPINGS            | 14    | direct_typings             | TYPING_START                                |
    +----------------------------------------+-------+----------------------------+---------------------------------------------+
    """
    __keys__ = {
        'guilds': INTENT_SHIFT_GUILDS,
        'guild_users': INTENT_SHIFT_GUILD_USERS,
        'guild_bans': INTENT_SHIFT_GUILD_BANS,
        'guild_emojis_and_stickers': INTENT_SHIFT_GUILD_EMOJIS_AND_STICKERS,
        'guild_integrations': INTENT_SHIFT_GUILD_INTEGRATIONS,
        'guild_webhooks': INTENT_SHIFT_GUILD_WEBHOOKS,
        'guild_invites': INTENT_SHIFT_GUILD_INVITES,
        'guild_voice_states': INTENT_SHIFT_GUILD_VOICE_STATES,
        'guild_presences': INTENT_SHIFT_GUILD_PRESENCES,
        'guild_messages': INTENT_SHIFT_GUILD_MESSAGES,
        'guild_reactions': INTENT_SHIFT_GUILD_REACTIONS,
        'guild_typings': INTENT_SHIFT_GUILD_TYPINGS,
        'direct_messages': INTENT_SHIFT_DIRECT_MESSAGES,
        'direct_reactions': INTENT_SHIFT_DIRECT_REACTIONS,
        'direct_typings': INTENT_SHIFT_DIRECT_TYPINGS,
    }
    
    def __new__(cls, int_=-1):
        """
        Creates a new ``IntentFlag`` instance from the passed `int_`. If any invalid intent flag is passed, those
        will be removed. If the wrapper is started up without presence caching, then `.guild_presences` will be
        set to `False` by default.
        
        Parameters
        ----------
        int_ : `int` instance, Optional
            The value what will be converted ``IntentFlag`` instance. If not passed or passed as a negative value,
            then returns an ``IntentFlag`` what contains all the enabled flags.
        
        Returns
        -------
        intent_flag : ``IntentFlag``
        
        Raises
        ------
        TypeError
            If `int_` was not passed as `int` instance.
        
        Notes
        -----
        The defaultly created intent flags contain the privileged gateway intents, so if you have those disabled, or
        if those are not allowed for you, then make sure, you specify them.
        """
        if not isinstance(int_, int):
            raise TypeError(f'{cls.__name__} expected `int` instance, got `{int_!r}')
        
        intent_flag = 0
        if int_ < 0:
            for value in cls.__keys__.values():
                intent_flag = intent_flag|(1<<value)
            
            # If presence cache is disabled, disable presence updates
            if not CACHE_PRESENCE:
                intent_flag = intent_flag^(1<<INTENT_SHIFT_GUILD_PRESENCES)
        else:
            for value in cls.__keys__.values():
                if (int_>>value)&1:
                    intent_flag = intent_flag|(1<<value)
            
            # If presence cache is disabled, disable presence updates
            if not CACHE_PRESENCE:
                if intent_flag&INTENT_MASK_GUILD_PRESENCES:
                    intent_flag = intent_flag^INTENT_MASK_GUILD_PRESENCES
        
        return int.__new__(cls, intent_flag)
    
    def iterate_parser_names(self):
        """
        Yields every parser's name, what the intent flag allows to be received.
        
        This method is a generator.
        
        Yields
        ------
        parser_name : `str`
        """
        for shift in self.__keys__.values():
            if (self>>shift)&1:
                yield from INTENT_SHIFT_EVENTS[shift]
        
        yield from GLOBAL_INTENT_SHIFT_EVENTS



__all__ = ('DiscordException',)


from ...backend.headers import RETRY_AFTER, DATE

from ..utils import parse_date_header_to_datetime, DATETIME_FORMAT_CODE

class DiscordException(Exception):
    """
    Represents an exception raised by Discord, when it response with a not expected response code.
    
    Depending on Discord's response code, the http client's behaviours differently.
    
    +-------+-----------------------+---------------+
    | Code  | Meaning               | Behaviour     |
    +=======+=======================+===============+
    | 200   | OK                    | return        |
    +-------+-----------------------+---------------+
    | 201   | CREATED               | return        |
    +-------+-----------------------+---------------+
    | 204   | NO CONTENT            | return        |
    +-------+-----------------------+---------------+
    | 304   | NOT MODIFIED          | return        |
    +-------+-----------------------+---------------+
    | 400   | BAD REQUEST           | raise         |
    +-------+-----------------------+---------------+
    | 401   | UNAUTHORIZED          | raise         |
    +-------+-----------------------+---------------+
    | 403   | FORBIDDEN             | raise         |
    +-------+-----------------------+---------------+
    | 404   | NOT FOUND             | raise         |
    +-------+-----------------------+---------------+
    | 405   | METHOD NOT ALLOWED    | raise         |
    +-------+-----------------------+---------------+
    | 429   | TOO MANY REQUESTS     | rate limited  |
    +-------+-----------------------+---------------+
    | 500   | SERVER ERROR          | \*retry       |
    +-------+-----------------------+---------------+
    | 502   | GATEWAY UNAVAILABLE   | \*retry       |
    +-------+-----------------------+---------------+
    | 5XX   | SERVER ERROR          | raise         |
    +-------+-----------------------+---------------+
    
    \* For five times a request can fail with `OsError` or return `501` / `502` response code. If the request fails
    with these cases for the fifth try and the last one resulted `501` or `502` response code, then
    ``DiscordException`` will be raised.
    
    Attributes
    ----------
    response : ``ClientResponse``
        The http client response, what caused the error.
    data : `Any`
        Deserialized `json` response data if applicable.
    _messages : `None` or `list` of `str`
        Initially the `._messages` attribute is `None`, but when the `.messages` property is used for the first time,
        the messages will be parsed out from the response and from it's data.
    _code : `None` or `int`
        Initially the `._code` attribute is set to `None`, but first time when the `.code` property is accessed, it is
        parsed out. If the response data does not contains `code`, then this attribute is set to `0`.
    """
    def __init__(self, response, data):
        """
        Creates a new ``DiscordException``.
        
        Parameters
        ----------
        response : ``ClientResponse``
            The http client response, what caused the error.
        data : `Any`
            Deserialized `json` response data if applicable.
        """
        Exception.__init__(self)
        self.response = response
        self.data = data
        self._messages = None
        self._code = None
    
    @property
    def messages(self):
        """
        Returns a list of the errors. The 0th element of the list is always a header line, what contains the
        exception's name, the response's reason and it's status. If set, then also the Discord's internal error code
        and it's message as well.
        
        Every other element at the list is optional. Those are extra errors included in the response's data.
        
        Returns
        -------
        messages : `list` of `str`
        """
        messages = self._messages
        if messages is None:
            messages = self._cr_messages()
        return messages
    
    def _cr_messages(self):
        """
        Generates the exception's messages from the causer response's headers. If the response's data contains `code`
        or / and `message` as well, then it will complement the exception message's header line with those too.
        
        If the response's data contains additional errors too, then those will be parsed out, and added to the list.
        
        Saves the result to the `._messages` instance attribute and saves it as well.
        
        Returns
        -------
        messages : `list` of `str`
        """
        messages = []
        
        message_parts = []
        data = self.data
        if type(data) is dict:
            message_base = data.get('message', '')
            error_datas = data.get('errors', None)
            if (error_datas is not None) and error_datas:
                stack = [[(None, error_datas,)]]
                while True:
                    line = stack[-1]
                    if not line:
                        del stack[-1]
                        if not stack:
                            break
                            
                        del stack[-1][-1]
                        if not message_parts:
                            continue
                            
                        del message_parts[-1]
                        if not message_parts:
                            continue
                        
                        if message_parts[-1] != '.':
                            continue
                        
                        del message_parts[-1]
                        continue
                    
                    key, value = line[-1]
                    
                    if type(value) is dict:
                        if (key is not None):
                            if key.isdigit():
                                # this should not be first ever
                                message_parts.append(f'[{key}]')
                            else:
                                if message_parts:
                                    message_parts.append('.')
                                message_parts.append(key)
                        try:
                            error_datas = value['_errors']
                        except KeyError:
                            stack.append(list(value.items()))
                            continue
                        
                        for error_data in error_datas:
                            error_data_length = len(error_data)
                            try:
                                error_code = error_data['code']
                            except KeyError:
                                error_code = 'ERROR'
                            else:
                                error_data_length -= 1
                            
                            try:
                                error_message = error_data['message']
                            except KeyError:
                                error_message = None
                            else:
                                error_data_length -= 1
                            
                            if error_data_length:
                                error_extra = ' '.join(
                                    f'{key}={value!r}' for key, value in error_data.items()
                                        if key not in ('code', 'message')
                                )
                                
                                if (error_message is None):
                                    error_message = error_extra
                                else:
                                    error_message = f'{error_message!r} {error_extra}'
                            
                            elif (error_message is None):
                                error_message = ''
                            else:
                                error_message = repr(error_message)
                            
                            if message_parts:
                                message_parts.append('.')
                            
                            message_parts.append(f'{error_code}({error_message})')
                            messages.append(''.join(message_parts))
                            del message_parts[-1]
                            if not message_parts:
                                continue
                            
                            if message_parts[-1] != '.':
                                continue
                            
                            del message_parts[-1]
                            continue
                        
                        del line[-1]
                        if not message_parts:
                            continue
                            
                        del message_parts[-1]
                        
                        if not message_parts:
                            continue
                            
                        if message_parts[-1] != '.':
                            continue
                        
                        del message_parts[-1]
                        continue
                    
                    if key.isdigit():
                        message_parts.append(f'[{key}]')
                    else:
                        message_parts.append('.')
                        message_parts.append(key)
                    message_parts.append('.')
                    message_parts.append(value)
                    messages.append(''.join(message_parts))
                    del line[-1]
                    del message_parts[-3:]
                    if not message_parts:
                        continue
                    
                    if message_parts[-1] != '.':
                        continue
                    
                    del message_parts[-1]
                    
        else:
            message_base = ''
        
        response = self.response
        message_parts.append(parse_date_header_to_datetime(response.headers[DATE]).__format__(DATETIME_FORMAT_CODE))
        message_parts.append('; ')
        message_parts.append(response.method)
        message_parts.append(' ')
        message_parts.append(str(response.url))
        messages.append(''.join(message_parts))
        
        message_parts.clear()
        
        message_parts.append(f'{self.__class__.__name__} {response.reason} ({response.status})')
        
        code = self.code
        if code:
            message_parts.append(f', code=')
            message_parts.append(repr(code))

        if message_base:
            message_parts.append(': ')
            message_parts.append(message_base)
        elif messages:
            message_parts.append(':')
        
        messages.append(''.join(message_parts))
        messages.reverse()
        
        if self.response.status == 429:
            try:
                retry_after = self.response.headers[RETRY_AFTER]
            except KeyError:
                pass
            else:
                messages.append(f'{RETRY_AFTER}: {retry_after}')
        
        self._messages = messages
        return messages
    
    def __repr__(self):
        """Returns the representation of the object."""
        return '\n'.join(self.messages)
    
    __str__ = __repr__
    
    @property
    def code(self):
        """
        Returns the Discord's internal exception code, if it is included in the response's data. If not, then returns
        `0`.
        
        Returns
        -------
        error_code : `int`
        """
        code = self._code
        if code is None:
            code = self._cr_code()
        return code
    
    def _cr_code(self):
        """
        Parses out the Discord's inner exception code from the response's data. Sets it to `._code` and returns it as
        well.
        
        Returns
        -------
        error_code : `int`
        """
        data = self.data
        if type(data) is dict:
            code = data.get('code', 0)
        else:
            code = 0
        
        self._code = code
        return code
    
    @property
    def status(self):
        """
        The exception's response's status.
        
        Returns
        -------
        status_code : `int`
        """
        return self.response.status
